"""The base hosting service class and associated definitions."""

from __future__ import unicode_literals

import base64
import hashlib
import json
import logging
import mimetools
import re
import ssl

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from django.conf.urls import include, url
from django.dispatch import receiver
from django.utils import six
from django.utils.six.moves.urllib.error import URLError
from django.utils.six.moves.urllib.parse import urlparse
from django.utils.six.moves.urllib.request import (Request as BaseURLRequest,
                                                   HTTPBasicAuthHandler,
                                                   urlopen)
from django.utils.translation import ugettext_lazy as _
from djblets.registries.errors import ItemLookupError
from djblets.registries.registry import (ALREADY_REGISTERED, LOAD_ENTRY_POINT,
                                         NOT_REGISTERED)

import reviewboard.hostingsvcs.urls as hostingsvcs_urls
from reviewboard.registries.registry import EntryPointRegistry
from reviewboard.scmtools.certs import Certificate
from reviewboard.scmtools.errors import UnverifiedCertificateError
from reviewboard.signals import initializing


class URLRequest(BaseURLRequest):
    """A request that can use any HTTP method.

    By default, the :py:class:`urllib2.Request` class only supports HTTP GET
    and HTTP POST methods. This subclass allows for any HTTP method to be
    specified for the request.
    """

    def __init__(self, url, body='', headers=None, method='GET'):
        """Initialize the URLRequest.

        Args:
            url (unicode):
                The URL to make the request against.

            body (unicode or bytes):
                The content of the request.

            headers (dict, optional):
                Additional headers to attach to the request.

            method (unicode, optional):
                The request method. If not provided, it defaults to a ``GET``
                request.
        """
        # Request is an old-style class and therefore we cannot use super().
        BaseURLRequest.__init__(self, url, body, headers or {})
        self.method = method

    def get_method(self):
        """Return the HTTP method of the request.

        Returns:
            unicode:
            The HTTP method of the request.
        """
        return self.method

    def add_basic_auth(self, username, password):
        """Add HTTP Basic Authentication headers to the request.

        Args:
            username (unicode or bytes):
                The username.

            password (unicode or bytes):
                The password.
        """
        if isinstance(username, six.text_type):
            username = username.encode('utf-8')

        if isinstance(password, six.text_type):
            password = password.encode('utf-8')

        auth = b'%s:%s' % (username, password)
        self.add_header(HTTPBasicAuthHandler.auth_header,
                        b'Basic %s' % base64.b64encode(auth))


class HostingServiceClient(object):
    """Client for communicating with a hosting service's API.

    This implementation includes abstractions for performing HTTP operations,
    and wrappers for those to interpret responses as JSON data.

    HostingService subclasses can also include an override of this class to add
    additional checking (such as GitHub's checking of rate limit headers), or
    add higher-level API functionality.
    """

    def __init__(self, hosting_service):
        """Initialize the client.

        This method is a no-op. Subclasses requiring access to the hosting
        service or account should override this method.

        Args:
            hosting_service (HostingService):
                The hosting service that is using this client.
        """
        self.hosting_service = hosting_service

    #
    # HTTP utility methods
    #

    def http_delete(self, url, headers=None, *args, **kwargs):
        """Perform an HTTP DELETE on the given URL.

        Args:
            url (unicode):
                The URL to perform the request on.

            headers (dict, optional):
                Extra headers to include with the request.

            *args (tuple):
                Additional positional arguments to pass to
                :py:meth:`http_request`.

            **kwargs (dict):
                Additional keyword arguments to pass to
                :py:meth:`http_request`.

        Returns:
            tuple:
            A tuple of:

            * The response body (:py:class:`bytes`).
            * The response headers (:py:class:`dict`).

        Raises:
            urllib2.HTTPError:
                When the HTTP request fails.

            urllib2.URLError:
                When there is an error communicating with the URL.
        """
        return self.http_request(url, headers=headers, method='DELETE',
                                 **kwargs)

    def http_get(self, url, headers=None, *args, **kwargs):
        """Perform an HTTP GET on the given URL.

        Args:
            url (unicode):
                The URL to perform the request on.

            headers (dict, optional):
                Extra headers to include with the request.

            *args (tuple):
                Additional positional arguments to pass to
                :py:meth:`http_request`.

            **kwargs (dict):
                Additional keyword arguments to pass to
                :py:meth:`http_request`.

        Returns:
            tuple:
            A tuple of:

            * The response body (:py:class:`bytes`)
            * The response headers (:py:class:`dict`)

        Raises:
            urllib2.HTTPError:
                When the HTTP request fails.

            urllib2.URLError:
                When there is an error communicating with the URL.
        """
        return self.http_request(url, headers=headers, method='GET', **kwargs)

    def http_post(self, url, body=None, fields=None, files=None,
                  content_type=None, headers=None, *args, **kwargs):
        """Perform an HTTP POST on the given URL.

        Args:
            url (unicode):
                The URL to perform the request on.

            body (unicode, optional):
                The request body. if not provided, it will be generated from
                the ``fields`` and ``files`` arguments.

            fields (dict, optional):
                Form fields to use to generate the request body. This argument
                will only be used if ``body`` is ``None``.

            files (dict, optional):
                Files to use to generate the request body. This argument will
                only be used if ``body`` is ``None``.

            content_type (unicode, optional):
                The content type of the request. If provided, it will be
                appended as the :mailheader:`Content-Type` header.

            headers (dict, optional):
                Extra headers to include with the request.

            *args (tuple):
                Additional positional arguments to pass to
                :py:meth:`http_request`.

            **kwargs (dict):
                Additional keyword arguments to pass to
                :py:meth:`http_request`.

        Returns:
            tuple:
            A tuple of:

            * The response body (:py:class:`bytes`)
            * The response headers (:py:class:`dict`)

        Raises:
            urllib2.HTTPError:
                When the HTTP request fails.

            urllib2.URLError:
                When there is an error communicating with the URL.
        """
        if headers:
            headers = headers.copy()
        else:
            headers = {}

        if body is None:
            if fields is not None:
                body, content_type = self.build_form_data(fields, files)
            else:
                body = ''

        if content_type:
            headers['Content-Type'] = content_type

        headers['Content-Length'] = '%d' % len(body)

        return self.http_request(url, body=body, headers=headers,
                                 method='POST', **kwargs)

    def http_request(self, url, body=None, headers=None, method='GET',
                     username=None, password=None):
        """Perform some HTTP operation on a given URL.

        If the ``username`` and ``password`` arguments are provided, the
        headers required for HTTP Basic Authentication will be added to
        the request.

        Args:
            url (unicode):
                The URL to open.

            body (unicode, optional):
                The request body.

            headers (dict, optional):
                Headers to include in the request.

            method (unicode, optional):
                The HTTP method to use to perform the request.

            username (unicode, optional):
                The username to use for HTTP Basic Authentication.

            password (unicode, optional):
                The password to use for HTTP Basic Authentication.

        Returns:
            tuple:
            A tuple of:

            * The response body (:py:class:`bytes`)
            * The response headers (:py:class:`dict`)

        Raises:
            urllib2.HTTPError:
                When the HTTP request fails.

            urllib2.URLError:
                When there is an error communicating with the URL.
        """
        request = URLRequest(url, body, headers, method=method)

        if username is not None and password is not None:
            request.add_basic_auth(username, password)

        try:
            if (self.hosting_service and
                'ssl_cert' in self.hosting_service.account.data):
                # create_default_context only exists in Python 2.7.9+. Using it
                # here should be fine, however, because accepting invalid or
                # self-signed certificates is only possible when running
                # against versions that have this (see the check for
                # create_default_context below).
                context = ssl.create_default_context()
                context.load_verify_locations(
                    cadata=self.hosting_service.account.data['ssl_cert'])
                context.check_hostname = False
            else:
                context = None

            response = urlopen(request, context=context)
            return response.read(), response.headers
        except URLError as e:
            if ('CERTIFICATE_VERIFY_FAILED' not in six.text_type(e) or
                not hasattr(ssl, 'create_default_context')):
                raise

            parts = urlparse(url)
            port = parts.port or 443

            cert_pem = ssl.get_server_certificate((parts.hostname, port))
            cert_der = ssl.PEM_cert_to_DER_cert(cert_pem)

            cert = x509.load_pem_x509_certificate(cert_pem.encode('ascii'),
                                                  default_backend())
            issuer = cert.issuer.get_attributes_for_oid(
                x509.oid.NameOID.COMMON_NAME)[0].value
            subject = cert.subject.get_attributes_for_oid(
                x509.oid.NameOID.COMMON_NAME)[0].value

            raise UnverifiedCertificateError(
                Certificate(
                    pem_data=cert_pem,
                    valid_from=cert.not_valid_before.isoformat(),
                    valid_until=cert.not_valid_after.isoformat(),
                    issuer=issuer,
                    hostname=subject,
                    fingerprint=hashlib.sha256(cert_der).hexdigest()))

    #
    # JSON utility methods
    #

    def json_delete(self, *args, **kwargs):
        """Perform an HTTP DELETE and interpret the results as JSON.

        Args:
            *args (tuple):
                Additional positional arguments to pass to
                :py:meth:`http_delete`.

            **kwargs (dict):
                Additional keyword arguments to pass to
                :py:meth:`http_delete`.

        Returns:
            tuple:
            A tuple of:

            * The JSON data (in the appropriate type)
            * The response headers (:py:class:`dict`)

        Raises:
            urllib2.HTTPError:
                When the HTTP request fails.

            urllib2.URLError:
                When there is an error communicating with the URL.
        """
        return self._do_json_method(self.http_delete, *args, **kwargs)

    def json_get(self, *args, **kwargs):
        """Perform an HTTP GET and interpret the results as JSON.

        Args:
            *args (tuple):
                Additional positional arguments to pass to
                :py:meth:`http_get`.

            **kwargs (dict):
                Additional keyword arguments to pass to
                :py:meth:`http_get`.

        Returns:
            tuple:
            A tuple of:

            * The JSON data (in the appropriate type)
            * The response headers (:py:class:`dict`)

        Raises:
            urllib2.HTTPError:
                When the HTTP request fails.

            urllib2.URLError:
                When there is an error communicating with the URL.
        """
        return self._do_json_method(self.http_get, *args, **kwargs)

    def json_post(self, *args, **kwargs):
        """Perform an HTTP POST and interpret the results as JSON.

        Args:
            *args (tuple):
                Additional positional arguments to pass to
                :py:meth:`http_post`.

            **kwargs (dict):
                Additional keyword arguments to pass to
                :py:meth:`http_post`.

        Returns:
            tuple:
            A tuple of:

            * The JSON data (in the appropriate type)
            * The response headers (:py:class:`dict`)

        Raises:
            urllib2.HTTPError:
                When the HTTP request fails.

            urllib2.URLError:
                When there is an error communicating with the URL.
        """
        return self._do_json_method(self.http_post, *args, **kwargs)

    def _do_json_method(self, method, *args, **kwargs):
        """Parse the result of an HTTP operation as JSON.

        Args:
            method (callable):
                The callable to use to execute the request.

            *args (tuple):
                Positional arguments to pass to ``method``.

            **kwargs (dict):
                Keyword arguments to pass to ``method``.

        Returns:
            tuple:
            A tuple of:

            * The JSON data (in the appropriate type)
            * The response headers (:py:class:`dict`)

        Raises:
            urllib2.HTTPError:
                When the HTTP request fails.

            urllib2.URLError:
                When there is an error communicating with the URL.
        """
        data, headers = method(*args, **kwargs)

        if data:
            data = json.loads(data)

        return data, headers

    #
    # Internal utilities
    #

    @staticmethod
    def build_form_data(fields, files=None):
        """Encode data for use in an HTTP POST.

        Args:
            fields (dict):
                A mapping of field names (:py:class:`unicode`) to values.

            files (dict, optional):
                A mapping of field names (:py:class:`unicode`) to files
                (:py:class:`dict`).

        Returns:
            tuple:
            A tuple of:

            * The request content (:py:class:`unicode`).
            * The request content type (:py:class:`unicode`).
        """
        boundary = HostingServiceClient._make_form_data_boundary()
        content_parts = []

        for key, value in sorted(six.iteritems(fields),
                                 key=lambda pair: pair[0]):
            if isinstance(key, six.text_type):
                key = key.encode('utf-8')

            if isinstance(value, six.text_type):
                value = value.encode('utf-8')

            content_parts.append(
                b'--%(boundary)s\r\n'
                b'Content-Disposition: form-data; name="%(key)s"\r\n'
                b'\r\n'
                b'%(value)s\r\n'
                % {
                    'boundary': boundary,
                    'key': key,
                    'value': value,
                }
            )

        if files:
            for key, data in sorted(six.iteritems(files),
                                    key=lambda pair: pair[0]['filename']):
                filename = data['filename']
                content = data['content']

                if isinstance(key, six.text_type):
                    key = key.encode('utf-8')

                if isinstance(filename, six.text_type):
                    filename = filename.encode('utf-8')

                if isinstance(content, six.text_type):
                    content = content.encode('utf-8')

                content_parts.append(
                    b'--%(boundary)s\r\n'
                    b'Content-Disposition: form-data; name="%(key)s";'
                    b' filename="%(filename)s"\r\n'
                    b'\r\n'
                    b'%(value)s\r\n'
                    % {
                        'boundary': boundary,
                        'key': key,
                        'filename': filename,
                        'value': content,
                    }
                )

        content_parts.append(b'--%s--' % boundary)

        content = b''.join(content_parts)
        content_type = b'multipart/form-data; boundary=%s' % boundary

        return content, content_type

    @staticmethod
    def _make_form_data_boundary():
        """Return a unique boundary to use for HTTP form data.

        Returns:
            bytes:
            The boundary for use in the form data.
        """
        return mimetools.choose_boundary()


class HostingService(object):
    """An interface to a hosting service for repositories and bug trackers.

    HostingService subclasses are used to more easily configure repositories
    and to make use of third party APIs to perform special operations not
    otherwise usable by generic repositories.

    A HostingService can specify forms for repository and bug tracker
    configuration.

    It can also provide a list of repository "plans" (such as public
    repositories, private repositories, or other types available to the hosting
    service), along with configuration specific to the plan. These plans will
    be available when configuring the repository.
    """

    name = None
    plans = None
    supports_bug_trackers = False
    supports_post_commit = False
    supports_repositories = False
    supports_ssh_key_association = False
    supports_two_factor_auth = False
    supports_list_remote_repositories = False
    has_repository_hook_instructions = False

    self_hosted = False
    repository_url_patterns = None

    client_class = HostingServiceClient

    #: Optional form used to configure authentication settings for an account.
    auth_form = None

    # These values are defaults that can be overridden in repository_plans
    # above.
    needs_authorization = False
    supported_scmtools = []
    form = None
    fields = []
    repository_fields = {}
    bug_tracker_field = None

    def __init__(self, account):
        """Initialize the hosting service.

        Args:
            account (reviewboard.hostingsvcs.models.HostingServiceAccount):
                The account to use with the service.
        """
        assert account
        self.account = account

        self.client = self.client_class(self)

    def is_authorized(self):
        """Return whether or not the account is currently authorized.

        An account may no longer be authorized if the hosting service
        switches to a new API that doesn't match the current authorization
        records. This function will determine whether the account is still
        considered authorized.

        Returns:
            bool:
            Whether or not the associated account is authorized.
        """
        return False

    def get_password(self):
        """Return the raw password for this hosting service.

        Not all hosting services provide this, and not all would need it.
        It's primarily used when building a Subversion client, or other
        SCMTools that still need direct access to the repository itself.

        Returns:
            unicode:
            The password.
        """
        return None

    def is_ssh_key_associated(self, repository, key):
        """Return whether or not the key is associated with the repository.

        If the given key is present amongst the hosting service's deploy keys
        for the given repository, then it is considered to be associated.

        Sub-classes should implement this when the hosting service supports
        SSH key association.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository the key must be associated with.

            key (paramiko.PKey):
                The key to check for association.

        Returns:
            bool:
            Whether or not the key is associated with the repository.

        Raises:
            reviewboard.hostingsvcs.errors.SSHKeyAssociationError:
                If an error occured during communication with the hosting
                service.
        """
        raise NotImplementedError

    def associate_ssh_key(self, repository, key):
        """Associate an SSH key with a given repository.

        Sub-classes should implement this when the hosting service supports
        SSH key association.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to associate the key with.

            key (paramiko.PKey):
                The key to add to the repository's list of deploy keys.

        Raises:
            reviewboard.hostingsvcs.errors.SSHKeyAssociationError:
                If an error occured during key association.
        """
        raise NotImplementedError

    def authorize(self, username, password, hosting_url, credentials,
                  two_factor_auth_code=None, local_site_name=None,
                  *args, **kwargs):
        """Authorize an account for the hosting service.

        Args:
            username (unicode):
                The username for the account.

            password (unicode):
                The password for the account.

            hosting_url (unicode):
                The hosting URL for the service, if self-hosted.

            credentials (dict):
                All credentials provided by the authentication form. This
                will contain the username, password, and anything else
                provided by that form.

            two_factor_auth_code (unicode, optional):
                The two-factor authentication code provided by the user.

            local_site_name (unicode, optional):
                The Local Site name, if any, that the account should be
                bound to.

            *args (tuple):
                Extra unused positional arguments.

            **kwargs (dict):
                Extra keyword arguments containing values from the
                repository's configuration.

        Raises:
            reviewboard.hostingsvcs.errors.AuthorizationError:
                The credentials provided were not valid.

            reviewboard.hostingsvcs.errors.TwoFactorAuthCodeRequiredError:
                A two-factor authentication code is required to authorize
                this account. The request must be retried with the same
                credentials and with the ``two_factor_auth_code`` parameter
                provided.
        """
        raise NotImplementedError

    def check_repository(self, path, username, password, scmtool_class,
                         local_site_name, *args, **kwargs):
        """Checks the validity of a repository configuration.

        This performs a check against the hosting service or repository
        to ensure that the information provided by the user represents
        a valid repository.

        This is passed in the repository details, such as the path and
        raw credentials, as well as the SCMTool class being used, the
        LocalSite's name (if any), and all field data from the
        HostingServiceForm as keyword arguments.

        Args:
            path (unicode):
                The repository URL.

            username (unicode):
                The username to use.

            password (unicode):
                The password to use.

            scmtool_class (type):
                The subclass of :py:class:`~reviewboard.scmtools.core.SCMTool`
                that should be used.

            local_site_name (unicode):
                The name of the local site associated with the repository, or
                ``None``.

            *args (tuple):
                Additional positional arguments, unique to each hosting
                service.

            **kwargs (dict):
                Additional keyword arguments, unique to each hosting service.

        Raises:
            reviewboard.hostingsvcs.errors.RepositoryError:
                The repository is not valid.
        """
        scmtool_class.check_repository(path, username, password,
                                       local_site_name)

    def get_file(self, repository, path, revision, *args, **kwargs):
        """Return the requested file.

        Files can only be returned from hosting services that support
        repositories.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to retrieve the file from.

            path (unicode):
                The file path.

            revision (unicode):
                The revision the file should be retrieved from.

            *args (tuple):
                Ignored positional arguments.

            **kwargs (dict):
                Additional keyword arguments to pass to the SCMTool.

        Returns:
            bytes:
            The contents of the file.

        Raises:
            NotImplementedError:
                If this hosting service does not support repositories.
        """
        if not self.supports_repositories:
            raise NotImplementedError

        return repository.get_scmtool().get_file(path, revision, **kwargs)

    def get_file_exists(self, repository, path, revision, *args, **kwargs):
        """Return whether or not the given path exists in the repository.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to check for file existence.

            path (unicode):
                The file path.

            revision (unicode):
                The revision to check for file existence.

            *args (tuple):
                Ignored positional arguments.

            **kwargs (dict):
                Additional keyword arguments to be passed to the SCMTool.

        Returns:
            bool:
            Whether or not the file exists at the given revision in the
            repository.

        Raises:
            NotImplementedError:
                If this hosting service does not support repositories.
        """
        if not self.supports_repositories:
            raise NotImplementedError

        return repository.get_scmtool().file_exists(path, revision, **kwargs)

    def get_branches(self, repository):
        """Return a list of all branches in the repositories.

        This should be implemented by subclasses, and is expected to return a
        list of Branch objects. One (and only one) of those objects should have
        the "default" field set to True.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository for which branches should be returned.

        Returns:
            list of reviewboard.scmtools.core.Branch:
            The branches.
        """
        raise NotImplementedError

    def get_commits(self, repository, branch=None, start=None):
        """Return a list of commits backward in history from a given point.

        This should be implemented by subclasses, and is expected to return a
        list of Commit objects (usually 30, but this is flexible depending on
        the limitations of the APIs provided.

        This can be called multiple times in succession using the "parent"
        field of the last entry as the start parameter in order to paginate
        through the history of commits in the repository.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to retrieve commits from.

            branch (unicode, optional):
                The branch to retrieve from. If this is not provided, the
                default branch will be used.

            start (unicode, optional):
                An optional starting revision. If this is not provided, the
                most recent commits will be returned.

        Returns:
            list of reviewboard.scmtools.core.Commit:
            The retrieved commits.
        """
        raise NotImplementedError

    def get_change(self, repository, revision):
        """Return an individual change.

        This method should be implemented by subclasses.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to get the change from.

            revision (unicode):
                The revision to retrieve.

        Returns:
            reviewboard.scmtools.core.Commit:
            The change.
        """
        raise NotImplementedError

    def get_remote_repositories(self, owner=None, owner_type=None,
                                filter_type=None, start=None, per_page=None,
                                **kwargs):
        """Return a list of remote repositories for the owner.

        This method should be implemented by subclasses.

        Args:
            owner (unicode, optional):
                The owner of the repositories. This is usually a username.

            owner_type (unicode, optional):
                A hosting service-specific indicator of what the owner is (such
                as a user or a group).

            filter_type (unicode, optional):
                Some hosting service-specific criteria to filter by.

            start (int, optional):
                The index to start at.

            per_page (int, optional):
                The number of results per page.

        Returns:
            reviewboard.hostingsvcs.utils.APIPaginator:
            A paginator for the returned repositories.
        """
        raise NotImplementedError

    def get_remote_repository(self, repository_id):
        """Return the remote repository for the ID.

        This method should be implemented by subclasses.

        Args:
            repository_id (unicode):
                The repository's identifier. This is unique to each hosting
                service.

        Returns:
            reviewboard.hostingsvcs.repository.RemoteRepository:
            The remote repository.

        Raises:
            django.core.excptions.ObjectDoesNotExist:
                If the remote repository does not exist.
        """
        raise NotImplementedError

    @classmethod
    def get_repository_fields(cls, username, hosting_url, plan, tool_name,
                              field_vars):
        """Return the repository fields based on the given plan and tool.

        If the ``plan`` argument is specified, that will be used to fill in
        some tool-specific field values. Otherwise they will be retrieved from
        the :py:class:`HostingService`'s defaults.

        Args:
            username (unicode):
                The username.

            hosting_url (unicode):
                The URL of the repository.

            plan (unicode):
                The name of the plan.

            tool_name (unicode):
                The :py:attr:`~reviewboard.scmtools.core.SCMTool.name` of the
                :py:class:`~reviewboard.scmtools.core.SCMTool`.

            field_vars (dict):
                The field values from the hosting service form.

        Returns:
            dict:
            The filled in repository fields.

        Raises:
            KeyError:
               The provided plan is not valid for the hosting service.
        """
        if not cls.supports_repositories:
            raise NotImplementedError

        # Grab the list of fields for population below. We have to do this
        # differently depending on whether or not this hosting service has
        # different repository plans.
        fields = cls.get_field(plan, 'repository_fields')

        new_vars = field_vars.copy()
        new_vars['hosting_account_username'] = username

        if cls.self_hosted:
            new_vars['hosting_url'] = hosting_url
            new_vars['hosting_domain'] = urlparse(hosting_url)[1]

        results = {}

        assert tool_name in fields

        for field, value in six.iteritems(fields[tool_name]):
            try:
                results[field] = value % new_vars
            except KeyError as e:
                logging.error('Failed to generate %s field for hosting '
                              'service %s using %s and %r: Missing key %s'
                              % (field, six.text_type(cls.name), value,
                                 new_vars, e),
                              exc_info=1)
                raise KeyError(
                    _('Internal error when generating %(field)s field '
                      '(Missing key "%(key)s"). Please report this.') % {
                        'field': field,
                        'key': e,
                    })

        return results

    def get_repository_hook_instructions(self, request, repository):
        """Return instructions for setting up incoming webhooks.

        Subclasses can override this (and set
        `has_repository_hook_instructions = True` on the subclass) to provide
        instructions that administrators can see when trying to configure an
        incoming webhook for the hosting service.

        Args:
            request (django.http.HttpRequest):
                The current HTTP request.

            repository (reviewboard.scmtools.models.Repository):
                The repository for webhook setup instructions.

        Returns:
            django.utils.text.SafeText:
            Rendered and escaped HTML for displaying to the user.
        """
        raise NotImplementedError

    @classmethod
    def get_bug_tracker_requires_username(cls, plan=None):
        """Return whether or not the bug tracker requires usernames.

        Args:
            plan (unicode, optional):
                The name of the plan associated with the account.

        Raises:
            NotImplementedError:
                If the hosting service does not support bug tracking.
        """
        if not cls.supports_bug_trackers:
            raise NotImplementedError

        return ('%(hosting_account_username)s' in
                cls.get_field(plan, 'bug_tracker_field', ''))

    @classmethod
    def get_bug_tracker_field(cls, plan, field_vars):
        """Return the bug tracker field for the given plan.

        Args:
            plan (unicode):
                The name of the plan associated with the account.

            field_vars (dict):
                The field values from the hosting service form.

        Returns:
            unicode:
            The value of the bug tracker field.

        Raises
            KeyError:
               The provided plan is not valid for the hosting service.
        """
        if not cls.supports_bug_trackers:
            raise NotImplementedError

        bug_tracker_field = cls.get_field(plan, 'bug_tracker_field')

        if not bug_tracker_field:
            return ''

        try:
            return bug_tracker_field % field_vars
        except KeyError as e:
            logging.error('Failed to generate %s field for hosting '
                          'service %s using %r: Missing key %s'
                          % (bug_tracker_field, six.text_type(cls.name),
                             field_vars, e),
                          exc_info=1)
            raise KeyError(
                _('Internal error when generating %(field)s field '
                  '(Missing key "%(key)s"). Please report this.') % {
                    'field': bug_tracker_field,
                    'key': e,
                })

    @classmethod
    def get_field(cls, plan, name, default=None):
        """Return the value of the field for the given plan.

        If ``plan`` is not ``None``, the field will be looked up in the plan
        configuration for the service. Otherwise the hosting service's default
        value will be used.

        Args:
            plan (unicode):
                The plan name.

            name (unicode):
                The field name.

            default (unicode, optional):
                A default value if the field is not present.

        Returns:
            unicode:
            The field value.
        """
        if cls.plans:
            assert plan

            for plan_name, info in cls.plans:
                if plan_name == plan and name in info:
                    return info[name]

        return getattr(cls, name, default)


_hostingsvcs_urlpatterns = {}


class HostingServiceRegistry(EntryPointRegistry):
    """A registry for managing hosting services."""

    entry_point = 'reviewboard.hosting_services'
    lookup_attrs = ['hosting_service_id']

    errors = {
        ALREADY_REGISTERED: _(
            '"%(item)s" is already a registered hosting service.'
        ),
        LOAD_ENTRY_POINT: _(
            'Unable to load repository hosting service %(entry_point)s: '
            '%(error)s.'
        ),
        NOT_REGISTERED: _(
            '"%(attr_value)s" is not a registered hosting service.'
        ),
    }

    def __init__(self):
        super(HostingServiceRegistry, self).__init__()
        self._url_patterns = {}

    def unregister(self, service):
        """Unregister a hosting service.

        This will remove all registered URLs that the hosting service has
        defined.

        Args:
            service (type):
                The
                :py:class:`~reviewboard.hostingsvcs.service.HostingService`
                subclass.
        """
        super(HostingServiceRegistry, self).unregister(service)

        if service.hosting_service_id in self._url_patterns:
            cls_urlpatterns = self._url_patterns[service.hosting_service_id]
            hostingsvcs_urls.dynamic_urls.remove_patterns(cls_urlpatterns)
            del self._url_patterns[service.hosting_service_id]

    def process_value_from_entry_point(self, entry_point):
        """Load the class from the entry point.

        The ``id`` attribute will be set on the class from the entry point's
        name.

        Args:
            entry_point (pkg_resources.EntryPoint):
                The entry point.

        Returns:
            type:
            The :py:class:`HostingService` subclass.
        """
        cls = entry_point.load()
        cls.hosting_service_id = entry_point.name
        return cls

    def register(self, service):
        """Register a hosting service.

        This also adds the URL patterns defined by the hosting service. If the
        hosting service has a :py:attr:`HostingService.repository_url_patterns`
        attribute that is non-``None``, they will be automatically added.

        Args:
            service (type):
                The :py:class:`HostingService` subclass.
        """
        super(HostingServiceRegistry, self).register(service)

        if service.repository_url_patterns:
            cls_urlpatterns = [
                url(r'^(?P<hosting_service_id>%s)/'
                    % re.escape(service.hosting_service_id),
                    include(service.repository_url_patterns)),
            ]

            self._url_patterns[service.hosting_service_id] = cls_urlpatterns
            hostingsvcs_urls.dynamic_urls.add_patterns(cls_urlpatterns)


_hosting_service_registry = HostingServiceRegistry()


def get_hosting_services():
    """Return the list of hosting services.

    Returns:
        list:
        The :py:class:`~reviewboard.hostingsvcs.service.HostingService`
        subclasses.
    """
    return list(_hosting_service_registry)


def get_hosting_service(name):
    """Return the hosting service with the given name.

    If the hosting service is not found, None will be returned.
    """
    try:
        return _hosting_service_registry.get('hosting_service_id', name)
    except ItemLookupError:
        return None


def register_hosting_service(name, cls):
    """Register a custom hosting service class.

    A name can only be registered once. A KeyError will be thrown if attempting
    to register a second time.

    Args:
        name (unicode):
            The name of the hosting service.

        cls (type):
            The hosting service class. This should be a subclass of
            :py:class:`~reviewboard.hostingsvcs.service.HostingService`.
    """
    cls.hosting_service_id = name
    _hosting_service_registry.register(cls)


def unregister_hosting_service(name):
    """Unregister a previously registered hosting service.

    Args:
        name (unicode):
            The name of the hosting service.
    """
    try:
        _hosting_service_registry.unregister_by_attr('hosting_service_id',
                                                     name)
    except ItemLookupError as e:
        logging.error('Failed to unregister unknown hosting service "%s"'
                      % name)
        raise e


@receiver(initializing, dispatch_uid='populate_hosting_services')
def _on_initializing(**kwargs):
    _hosting_service_registry.populate()



"""Unit tests for the Codebase HQ hosting service."""

from __future__ import unicode_literals

from django.utils import six

from reviewboard.hostingsvcs.errors import RepositoryError
from reviewboard.hostingsvcs.testing import HostingServiceTestCase
from reviewboard.scmtools.crypto_utils import (decrypt_password,
                                               encrypt_password)
from reviewboard.scmtools.errors import FileNotFoundError


class CodebaseHQTests(HostingServiceTestCase):
    """Unit tests for the Codebase HQ hosting service."""

    service_name = 'codebasehq'
    fixtures = ['test_scmtools']

    default_account_data = {
        'domain': 'mydomain',
        'api_key': encrypt_password('abc123'),
        'password': encrypt_password(HostingServiceTestCase.default_password),
    }

    default_repository_extra_data = {
        'codebasehq_project_name': 'myproj',
        'codebasehq_repo_name': 'myrepo',
    }

    def test_service_support(self):
        """Testing CodebaseHQ service support capabilities"""
        self.assertFalse(self.service_class.supports_post_commit)
        self.assertTrue(self.service_class.supports_bug_trackers)
        self.assertTrue(self.service_class.supports_repositories)

    def test_get_repository_fields_for_git(self):
        """Testing CodebaseHQ.get_repository_fields for Git"""
        self.assertEqual(
            self.get_repository_fields(
                'Git',
                fields={
                    'codebasehq_project_name': 'myproj',
                    'codebasehq_repo_name': 'myrepo',
                }
            ),
            {
                'path': 'git@codebasehq.com:mydomain/myproj/myrepo.git',
            })

    def test_get_repository_fields_for_mercurial(self):
        """Testing CodebaseHQ.get_repository_fields for Mercurial"""
        self.assertEqual(
            self.get_repository_fields(
                'Mercurial',
                fields={
                    'codebasehq_project_name': 'myproj',
                    'codebasehq_repo_name': 'myrepo',
                }
            ),
            {
                'path': ('https://mydomain.codebasehq.com/projects/'
                         'myproj/repositories/myrepo/'),
            })

    def test_get_repository_fields_for_subversion(self):
        """Testing CodebaseHQ.get_repository_fields for Subversion"""
        self.assertEqual(
            self.get_repository_fields(
                'Subversion',
                fields={
                    'codebasehq_project_name': 'myproj',
                    'codebasehq_repo_name': 'myrepo',
                }
            ),
            {
                'path': 'https://mydomain.codebasehq.com/myproj/myrepo.svn',
            })

    def test_get_bug_tracker_field(self):
        """Testing CodebaseHQ.get_bug_tracker_field"""
        self.assertEqual(
            self.service_class.get_bug_tracker_field(
                None,
                {
                    'codebasehq_project_name': 'myproj',
                    'domain': 'mydomain',
                }),
            'https://mydomain.codebasehq.com/projects/myproj/tickets/%s')

    def test_check_repository_git(self):
        """Testing CodebaseHQ.check_repository for Git"""
        self._test_check_repository(codebase_scm_type='git',
                                    tool_name='Git')

    def test_check_repository_mercurial(self):
        """Testing CodebaseHQ.check_repository for Mercurial"""
        self._test_check_repository(codebase_scm_type='hg',
                                    tool_name='Mercurial')

    def test_check_repository_subversion(self):
        """Testing CodebaseHQ.check_repository for Subversion"""
        self._test_check_repository(codebase_scm_type='svn',
                                    tool_name='Subversion')

    def test_check_repository_with_mismatching_type(self):
        """Testing CodebaseHQ.check_repository with mismatching repository type
        """
        self._test_check_repository(codebase_scm_type='svn',
                                    tool_name='Mercurial',
                                    expect_success=False,
                                    expected_name_for_error='Subversion')

    def test_authorize(self):
        """Testing CodebaseHQ.authorize"""
        hosting_account = self.create_hosting_account(data={})

        with self.setup_http_test(payload=b'{}',
                                  hosting_account=hosting_account,
                                  expected_http_calls=1) as ctx:
            self.assertFalse(ctx.service.is_authorized())

            ctx.service.authorize(
                username='myuser',
                password='mypass',
                credentials={
                    'domain': 'mydomain',
                    'api_key': 'abc123',
                })

        ctx.assertHTTPCall(
            0,
            url='https://api3.codebasehq.com/users/myuser/public_keys',
            username='mydomain/myuser',
            password='abc123',
            headers={
                'Accept': 'application/xml',
            })

        self.assertEqual(set(six.iterkeys(hosting_account.data)),
                         {'api_key', 'domain', 'password'})
        self.assertEqual(decrypt_password(hosting_account.data['api_key']),
                         'abc123')
        self.assertEqual(hosting_account.data['domain'], 'mydomain')
        self.assertEqual(decrypt_password(hosting_account.data['password']),
                         'mypass')
        self.assertTrue(ctx.service.is_authorized())

    def test_get_file_with_mercurial(self):
        """Testing CodebaseHQ.get_file with Mercurial"""
        self._test_get_file(tool_name='Mercurial')

    def test_get_file_with_mercurial_not_found(self):
        """Testing CodebaseHQ.get_file with Mercurial with file not found"""
        self._test_get_file(tool_name='Mercurial', file_exists=False)

    def test_get_file_with_git(self):
        """Testing CodebaseHQ.get_file with Git"""
        self._test_get_file(tool_name='Git', expect_git_blob_url=True)

    def test_get_file_with_git_not_found(self):
        """Testing CodebaseHQ.get_file with Git with file not found"""
        self._test_get_file(tool_name='Git', expect_git_blob_url=True,
                            file_exists=False)

    def test_get_file_with_subversion(self):
        """Testing CodebaseHQ.get_file with Subversion"""
        self._test_get_file(tool_name='Subversion')

    def test_get_file_with_subversion_not_found(self):
        """Testing CodebaseHQ.get_file with Subversion with file not found"""
        self._test_get_file(tool_name='Subversion', file_exists=False)

    def test_get_file_exists_with_mercurial(self):
        """Testing CodebaseHQ.get_file_exists with Mercurial"""
        self._test_get_file_exists(tool_name='Mercurial')

    def test_get_file_exists_with_mercurial_not_found(self):
        """Testing CodebaseHQ.get_file_exists with Mercurial with file not
        found
        """
        self._test_get_file_exists(tool_name='Mercurial', file_exists=False)

    def test_get_file_exists_with_git(self):
        """Testing CodebaseHQ.get_file_exists with Git"""
        self._test_get_file_exists(tool_name='Git', expect_git_blob_url=True)

    def test_get_file_exists_with_git_not_found(self):
        """Testing CodebaseHQ.get_file_exists with Git with file not found"""
        self._test_get_file_exists(tool_name='Git', expect_git_blob_url=True,
                                   file_exists=False)

    def test_get_file_exists_with_subversion(self):
        """Testing CodebaseHQ.get_file_exists with Subversion"""
        self._test_get_file_exists(tool_name='Subversion')

    def test_get_file_exists_with_subversion_not_found(self):
        """Testing CodebaseHQ.get_file_exists with Subversion with file not
        found
        """
        self._test_get_file_exists(tool_name='Subversion', file_exists=False)

    def _test_check_repository(self, codebase_scm_type, tool_name,
                               expect_success=True,
                               expected_name_for_error=None):
        """Test repository checks.

        Args:
            codebase_scm_type (unicode):
                The name of the SCM type in the CodebaseHQ API to return in
                payloads.

            tool_name (unicode):
                The name of the SCM Tool to test with.

            expect_success (bool, optional):
                Whether to simulate a truthy response.

            expected_name_for_error (unicode, optional):
                The name of the SCM Tool to expect in the error response,
                if ``expect_success`` is ``False``.
        """
        payload = (
            '<?xml version="1.0" encoding="UTF-8"?>\n'
            '<repository>\n'
            ' <scm>%s</scm>\n'
            '</repository>\n'
            % codebase_scm_type
        ).encode('utf-8')

        check_repository_kwargs = {
            'codebasehq_project_name': 'myproj',
            'codebasehq_repo_name': 'myrepo',
            'tool_name': tool_name,
        }

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=1) as ctx:
            if expect_success:
                ctx.service.check_repository(**check_repository_kwargs)
            else:
                message = (
                    "The repository type doesn't match what you selected. Did "
                    "you mean %s?"
                    % expected_name_for_error
                )

                with self.assertRaisesMessage(RepositoryError, message):
                    ctx.service.check_repository(**check_repository_kwargs)

        ctx.assertHTTPCall(
            0,
            url='https://api3.codebasehq.com/myproj/myrepo',
            username='mydomain/myuser',
            password='abc123',
            headers={
                'Accept': 'application/xml',
            })

    def _test_get_file(self, tool_name, expect_git_blob_url=False,
                       file_exists=True):
        """Test file fetching.

        Args:
            tool_name (unicode):
                The name of the SCM Tool to test with.

            expect_git_blob_url (bool, optional):
                Whether to expect a URL referencing a Git blob.

            file_exists (bool, optional):
                Whether to simulate a truthy response.
        """
        if expect_git_blob_url:
            expected_url = 'https://api3.codebasehq.com/myproj/myrepo/blob/123'
        else:
            expected_url = \
                'https://api3.codebasehq.com/myproj/myrepo/blob/123/myfile'

        if file_exists:
            payload = b'My data\n'
            status_code = None
        else:
            payload = b''
            status_code = 404

        with self.setup_http_test(payload=payload,
                                  status_code=status_code,
                                  expected_http_calls=1) as ctx:
            repository = ctx.create_repository(tool_name=tool_name)

            get_file_kwargs = {
                'repository': repository,
                'path': 'myfile',
                'revision': '123',
            }

            if file_exists:
                result = ctx.service.get_file(**get_file_kwargs)

                self.assertIsInstance(result, bytes)
                self.assertEqual(result, b'My data\n')
            else:
                with self.assertRaises(FileNotFoundError):
                    ctx.service.get_file(**get_file_kwargs)

        ctx.assertHTTPCall(
            0,
            url=expected_url,
            username='mydomain/myuser',
            password='abc123',
            headers={
                'Accept': 'application/xml',
            })

    def _test_get_file_exists(self, tool_name, expect_git_blob_url=False,
                              file_exists=True):
        """Test file existence checks.

        Args:
            tool_name (unicode):
                The name of the SCM Tool to test with.

            expect_git_blob_url (bool, optional):
                Whether to expect a URL referencing a Git blob.

            file_exists (bool, optional):
                Whether to simulate a truthy response.
        """
        if file_exists:
            payload = b'{"scm": "git"}'
            status_code = None
        else:
            payload = None
            status_code = 404

        with self.setup_http_test(payload=payload,
                                  status_code=status_code,
                                  expected_http_calls=1) as ctx:
            repository = ctx.create_repository(tool_name=tool_name)
            result = ctx.service.get_file_exists(repository=repository,
                                                 path='myfile',
                                                 revision='123')

            self.assertEqual(result, file_exists)

        if expect_git_blob_url:
            expected_url = 'https://api3.codebasehq.com/myproj/myrepo/blob/123'
        else:
            expected_url = \
                'https://api3.codebasehq.com/myproj/myrepo/blob/123/myfile'

        ctx.assertHTTPCall(
            0,
            url=expected_url,
            username='mydomain/myuser',
            password='abc123',
            body=None,
            headers={
                'Accept': 'application/xml',
            })



from __future__ import unicode_literals

from django.contrib.auth.models import AnonymousUser, User
from django.template import Context
from django.test.client import RequestFactory
from django.utils import six
from djblets.testing.decorators import add_fixtures
from mock import Mock

from reviewboard.reviews.actions import (BaseReviewRequestAction,
                                         BaseReviewRequestMenuAction,
                                         MAX_DEPTH_LIMIT,
                                         clear_all_actions,
                                         get_top_level_actions,
                                         register_actions,
                                         unregister_actions)
from reviewboard.reviews.default_actions import (AddGeneralCommentAction,
                                                 CloseMenuAction,
                                                 DeleteAction,
                                                 DownloadDiffAction,
                                                 EditReviewAction,
                                                 ShipItAction,
                                                 SubmitAction,
                                                 UpdateMenuAction,
                                                 UploadDiffAction)
from reviewboard.reviews.errors import DepthLimitExceededError
from reviewboard.reviews.models import ReviewRequest
from reviewboard.testing import TestCase


class FooAction(BaseReviewRequestAction):
    action_id = 'foo-action'
    label = 'Foo Action'


class BarAction(BaseReviewRequestMenuAction):
    def __init__(self, action_id, child_actions=None):
        super(BarAction, self).__init__(child_actions)

        self.action_id = 'bar-' + action_id


class TopLevelMenuAction(BaseReviewRequestMenuAction):
    action_id = 'top-level-menu-action'
    label = 'Top Level Menu Action'


class PoorlyCodedAction(BaseReviewRequestAction):
    def get_label(self, context):
        raise Exception


class ActionsTestCase(TestCase):
    """Test case for unit tests dealing with actions."""

    def tearDown(self):
        super(ActionsTestCase, self).tearDown()

        # This prevents registered/unregistered/modified actions from leaking
        # between different unit tests.
        clear_all_actions()

    def make_nested_actions(self, depth):
        """Return a nested list of actions to register.

        This returns a list of actions, each entry nested within the prior
        entry, of the given length. The resulting list is intended to be
        registered.

        Args:
            depth (int):
                The nested depth for the actions.

        Returns:
            list of reviewboard.reviews.actions.BaseReviewRequestAction:
            The list of actions.
        """
        actions = [None] * depth
        actions[0] = BarAction('0')

        for i in range(1, depth):
            actions[i] = BarAction(six.text_type(i), [actions[i - 1]])

        return actions


class ActionRegistryTests(ActionsTestCase):
    """Unit tests for the review request actions registry."""

    def test_register_actions_with_invalid_parent_id(self):
        """Testing register_actions with an invalid parent ID"""
        foo_action = FooAction()

        message = (
            'bad-id does not correspond to a registered review request action'
        )

        foo_action.register()

        with self.assertRaisesMessage(KeyError, message):
            register_actions([foo_action], 'bad-id')

    def test_register_actions_with_already_registered_action(self):
        """Testing register_actions with an already registered action"""
        foo_action = FooAction()

        message = (
            '%s already corresponds to a registered review request action'
            % foo_action.action_id
        )

        foo_action.register()

        with self.assertRaisesMessage(KeyError, message):
            register_actions([foo_action])

    def test_register_actions_with_max_depth(self):
        """Testing register_actions with max_depth"""
        actions = self.make_nested_actions(MAX_DEPTH_LIMIT)
        extra_action = BarAction('extra')
        foo_action = FooAction()

        for d, action in enumerate(actions):
            self.assertEquals(action.max_depth, d)

        register_actions([extra_action], actions[0].action_id)
        actions = [extra_action] + actions

        for d, action in enumerate(actions):
            self.assertEquals(action.max_depth, d)

        register_actions([foo_action])
        self.assertEquals(foo_action.max_depth, 0)

    def test_register_actions_with_too_deep(self):
        """Testing register_actions with exceeding max depth"""
        actions = self.make_nested_actions(MAX_DEPTH_LIMIT + 1)
        invalid_action = BarAction(str(len(actions)), [actions[-1]])

        error_message = (
            '%s exceeds the maximum depth limit of %d'
            % (invalid_action.action_id, MAX_DEPTH_LIMIT)
        )

        with self.assertRaisesMessage(DepthLimitExceededError, error_message):
            register_actions([invalid_action])

    def test_unregister_actions(self):
        """Testing unregister_actions"""

        orig_action_ids = {
            action.action_id
            for action in get_top_level_actions()
        }
        self.assertIn('update-review-request-action', orig_action_ids)
        self.assertIn('review-action', orig_action_ids)

        unregister_actions(['update-review-request-action', 'review-action'])

        new_action_ids = {
            action.action_id
            for action in get_top_level_actions()
        }
        self.assertEqual(len(orig_action_ids), len(new_action_ids) + 2)
        self.assertNotIn('update-review-request-action', new_action_ids)
        self.assertNotIn('review-action', new_action_ids)

    def test_unregister_actions_with_child_action(self):
        """Testing unregister_actions with child action"""
        menu_action = TopLevelMenuAction([
            FooAction()
        ])

        self.assertEqual(len(menu_action.child_actions), 1)
        unregister_actions([FooAction.action_id])
        self.assertEqual(len(menu_action.child_actions), 0)

    def test_unregister_actions_with_unregistered_action(self):
        """Testing unregister_actions with unregistered action"""
        foo_action = FooAction()
        error_message = (
            '%s does not correspond to a registered review request action'
            % foo_action.action_id
        )

        with self.assertRaisesMessage(KeyError, error_message):
            unregister_actions([foo_action.action_id])

    def test_unregister_actions_with_max_depth(self):
        """Testing unregister_actions with max_depth"""
        actions = self.make_nested_actions(MAX_DEPTH_LIMIT + 1)

        unregister_actions([actions[0].action_id])
        extra_action = BarAction(str(len(actions)), [actions[-1]])
        extra_action.register()
        self.assertEquals(extra_action.max_depth, MAX_DEPTH_LIMIT)


class BaseReviewRequestActionTests(ActionsTestCase):
    """Unit tests for BaseReviewRequestAction."""

    def test_register_then_unregister(self):
        """Testing BaseReviewRequestAction.register then unregister for
        actions
        """
        foo_action = FooAction()
        foo_action.register()

        self.assertIn(foo_action.action_id, (
            action.action_id
            for action in get_top_level_actions()
        ))

        foo_action.unregister()

        self.assertNotIn(foo_action.action_id, (
            action.action_id
            for action in get_top_level_actions()
        ))

    def test_register_with_already_registered(self):
        """Testing BaseReviewRequestAction.register with already registered
        action
        """
        foo_action = FooAction()
        error_message = (
            '%s already corresponds to a registered review request action'
            % foo_action.action_id
        )

        foo_action.register()

        with self.assertRaisesMessage(KeyError, error_message):
            foo_action.register()

    def test_register_with_too_deep(self):
        """Testing BaseReviewRequestAction.register with exceeding max depth"""
        actions = self.make_nested_actions(MAX_DEPTH_LIMIT + 1)
        invalid_action = BarAction(str(len(actions)), [actions[-1]])
        error_message = (
            '%s exceeds the maximum depth limit of %d'
            % (invalid_action.action_id, MAX_DEPTH_LIMIT)
        )

        with self.assertRaisesMessage(DepthLimitExceededError, error_message):
            invalid_action.register()

    def test_unregister_with_unregistered_action(self):
        """Testing BaseReviewRequestAction.unregister with unregistered
        action
        """
        foo_action = FooAction()

        message = (
            '%s does not correspond to a registered review request action'
            % foo_action.action_id
        )

        with self.assertRaisesMessage(KeyError, message):
            foo_action.unregister()

    def test_unregister_with_max_depth(self):
        """Testing BaseReviewRequestAction.unregister with max_depth"""
        actions = self.make_nested_actions(MAX_DEPTH_LIMIT + 1)
        actions[0].unregister()
        extra_action = BarAction(str(len(actions)), [actions[-1]])

        extra_action.register()
        self.assertEquals(extra_action.max_depth, MAX_DEPTH_LIMIT)

    def test_init_already_registered_in_menu(self):
        """Testing BaseReviewRequestAction.__init__ for already registered
        action when nested in a menu action
        """
        foo_action = FooAction()
        error_message = ('%s already corresponds to a registered review '
                         'request action') % foo_action.action_id

        foo_action.register()

        with self.assertRaisesMessage(KeyError, error_message):
            TopLevelMenuAction([
                foo_action,
            ])

    def test_render_pops_context_even_after_error(self):
        """Testing BaseReviewRequestAction.render pops the context after an
        error
        """
        context = Context({'comment': 'this is a comment'})
        old_dict_count = len(context.dicts)
        poorly_coded_action = PoorlyCodedAction()

        with self.assertRaises(Exception):
            poorly_coded_action.render(context)

        new_dict_count = len(context.dicts)
        self.assertEquals(old_dict_count, new_dict_count)


class AddGeneralCommentActionTests(ActionsTestCase):
    """Unit tests for AddGeneralCommentAction."""

    fixtures = ['test_users']

    def setUp(self):
        super(AddGeneralCommentActionTests, self).setUp()

        self.action = AddGeneralCommentAction()

    def test_should_render_with_authenticated(self):
        """Testing AddGeneralCommentAction.should_render with authenticated
        user
        """
        request = RequestFactory().request()
        request.user = User.objects.get(username='doc')

        self.assertTrue(self.action.should_render({'request': request}))

    def test_should_render_with_anonymous(self):
        """Testing AddGeneralCommentAction.should_render with authenticated
        user
        """
        request = RequestFactory().request()
        request.user = AnonymousUser()

        self.assertFalse(self.action.should_render({'request': request}))


class CloseMenuActionTests(ActionsTestCase):
    """Unit tests for CloseMenuAction."""

    fixtures = ['test_users']

    def setUp(self):
        super(CloseMenuActionTests, self).setUp()

        self.action = CloseMenuAction()

    def test_should_render_for_owner(self):
        """Testing CloseMenuAction.should_render for owner of review request"""
        review_request = self.create_review_request(publish=True)

        request = RequestFactory().request()
        request.user = review_request.submitter

        self.assertTrue(self.action.should_render({
            'review_request': review_request,
            'request': request,
            'perms': {
                'reviews': {
                    'can_change_status': False,
                },
            },
        }))

    def test_should_render_for_owner_unpublished(self):
        """Testing CloseMenuAction.should_render for owner of review
        unpublished review request
        """
        review_request = self.create_review_request(public=False)

        request = RequestFactory().request()
        request.user = review_request.submitter

        self.assertTrue(self.action.should_render({
            'review_request': review_request,
            'request': request,
            'perms': {
                'reviews': {
                    'can_change_status': False,
                },
            },
        }))

    def test_should_render_for_user(self):
        """Testing CloseMenuAction.should_render for normal user"""
        review_request = self.create_review_request(publish=True)

        request = RequestFactory().request()
        request.user = User.objects.create_user(username='test-user',
                                                email='user@example.com')

        self.assertFalse(self.action.should_render({
            'review_request': review_request,
            'request': request,
            'perms': {
                'reviews': {
                    'can_change_status': False,
                },
            },
        }))

    def test_should_render_user_with_can_change_status(self):
        """Testing CloseMenuAction.should_render for user with
        can_change_status permission
        """
        review_request = self.create_review_request(publish=True)

        request = RequestFactory().request()
        request.user = User.objects.create_user(username='test-user',
                                                email='user@example.com')

        self.assertTrue(self.action.should_render({
            'review_request': review_request,
            'request': request,
            'perms': {
                'reviews': {
                    'can_change_status': True,
                },
            },
        }))

    def test_should_render_user_with_can_change_status_and_unpublished(self):
        """Testing CloseMenuAction.should_render for user with
        can_change_status permission and unpublished review request
        """
        review_request = self.create_review_request(public=False)

        request = RequestFactory().request()
        request.user = User.objects.create_user(username='test-user',
                                                email='user@example.com')

        self.assertFalse(self.action.should_render({
            'review_request': review_request,
            'request': request,
            'perms': {
                'reviews': {
                    'can_change_status': True,
                },
            },
        }))

    def test_should_render_with_discarded(self):
        """Testing CloseMenuAction.should_render with discarded review request
        """
        review_request = \
            self.create_review_request(status=ReviewRequest.DISCARDED)

        request = RequestFactory().request()
        request.user = review_request.submitter

        self.assertFalse(self.action.should_render({
            'review_request': review_request,
            'request': request,
            'perms': {
                'reviews': {
                    'can_change_status': False,
                },
            },
        }))

    def test_should_render_with_submitted(self):
        """Testing CloseMenuAction.should_render with submitted review request
        """
        review_request = \
            self.create_review_request(status=ReviewRequest.SUBMITTED)

        request = RequestFactory().request()
        request.user = review_request.submitter

        self.assertFalse(self.action.should_render({
            'review_request': review_request,
            'request': request,
            'perms': {
                'reviews': {
                    'can_change_status': False,
                },
            },
        }))


class DeleteActionTests(ActionsTestCase):
    """Unit tests for DeleteAction."""

    def setUp(self):
        super(DeleteActionTests, self).setUp()

        self.action = DeleteAction()

    def test_should_render_with_published(self):
        """Testing DeleteAction.should_render with standard user"""
        self.assertFalse(self.action.should_render({
            'perms': {
                'reviews': {
                    'delete_reviewrequest': False,
                },
            },
        }))

    def test_should_render_with_permission(self):
        """Testing SubmitAction.should_render with delete_reviewrequest
        permission
        """
        self.assertTrue(self.action.should_render({
            'perms': {
                'reviews': {
                    'delete_reviewrequest': True,
                },
            },
        }))


class DownloadDiffActionTests(ActionsTestCase):
    """Unit tests for DownloadDiffAction."""

    fixtures = ['test_users']

    def setUp(self):
        super(DownloadDiffActionTests, self).setUp()

        self.action = DownloadDiffAction()

    def test_get_url_on_diff_viewer(self):
        """Testing DownloadDiffAction.get_url on diff viewer page"""
        request = RequestFactory().request()
        request.resolver_match = Mock()
        request.resolver_match.url_name = 'view-diff'

        self.assertEqual(self.action.get_url({'request': request}),
                         'raw/')

    def test_get_url_on_interdiff(self):
        """Testing DownloadDiffAction.get_url on diff viewer interdiff page"""
        request = RequestFactory().request()
        request.resolver_match = Mock()
        request.resolver_match.url_name = 'view-interdiff'

        self.assertEqual(self.action.get_url({'request': request}),
                         'raw/')

    def test_get_url_on_diff_viewer_revision(self):
        """Testing DownloadDiffAction.get_url on diff viewer revision page"""
        request = RequestFactory().request()
        request.resolver_match = Mock()
        request.resolver_match.url_name = 'view-diff-revision'

        self.assertEqual(self.action.get_url({'request': request}),
                         'raw/')

    def test_get_url_on_review_request(self):
        """Testing DownloadDiffAction.get_url on review request page"""
        request = RequestFactory().request()
        request.resolver_match = Mock()
        request.resolver_match.url_name = 'review-request-detail'

        review_request = self.create_review_request()

        self.assertEqual(
            self.action.get_url({
                'request': request,
                'review_request': review_request,
            }),
            '/r/%s/diff/raw/' % review_request.display_id)

    @add_fixtures(['test_site'])
    def test_get_url_on_review_request_with_local_site(self):
        """Testing DownloadDiffAction.get_url on review request page with
        LocalSite
        """
        request = RequestFactory().request()
        request.resolver_match = Mock()
        request.resolver_match.url_name = 'review-request-detail'
        request._local_site_name = self.local_site_name

        review_request = self.create_review_request(id=123,
                                                    with_local_site=True)

        self.assertEqual(
            self.action.get_url({
                'request': request,
                'review_request': review_request,
            }),
            '/s/%s/r/%s/diff/raw/' % (self.local_site_name,
                                      review_request.display_id))

    def test_get_hidden_on_diff_viewer(self):
        """Testing DownloadDiffAction.get_hidden on diff viewer page"""
        request = RequestFactory().request()
        request.resolver_match = Mock()
        request.resolver_match.url_name = 'view-diff'

        self.assertFalse(self.action.get_hidden({'request': request}))

    def test_get_hidden_on_interdiff(self):
        """Testing DownloadDiffAction.get_hidden on diff viewer interdiff page
        """
        request = RequestFactory().request()
        request.resolver_match = Mock()
        request.resolver_match.url_name = 'view-interdiff'

        self.assertTrue(self.action.get_hidden({'request': request}))

    def test_get_hidden_on_diff_viewer_revision(self):
        """Testing DownloadDiffAction.get_hdiden on diff viewer revision page
        """
        request = RequestFactory().request()
        request.resolver_match = Mock()
        request.resolver_match.url_name = 'view-diff-revision'

        self.assertFalse(self.action.get_hidden({'request': request}))

    def test_get_hidden_on_review_request(self):
        """Testing DownloadDiffAction.get_hdiden on diff viewer revision page
        """
        request = RequestFactory().request()
        request.resolver_match = Mock()
        request.resolver_match.url_name = 'review-request-detail'

        review_request = self.create_review_request()

        self.assertFalse(self.action.get_hidden({
            'request': request,
            'review_request': review_request,
        }))

    def test_should_render_on_diff_viewer(self):
        """Testing DownloadDiffAction.should_render on diff viewer page"""
        request = RequestFactory().request()
        request.resolver_match = Mock()
        request.resolver_match.url_name = 'view-diff'

        review_request = self.create_review_request()

        self.assertTrue(self.action.should_render({
            'request': request,
            'review_request': review_request,
        }))

    def test_should_render_on_interdiff(self):
        """Testing DownloadDiffAction.should_render on diff viewer interdiff
        page
        """
        request = RequestFactory().request()
        request.resolver_match = Mock()
        request.resolver_match.url_name = 'view-interdiff'

        review_request = self.create_review_request()

        self.assertTrue(self.action.should_render({
            'request': request,
            'review_request': review_request,
        }))

    def test_should_render_on_diff_viewer_revision(self):
        """Testing DownloadDiffAction.should_render on diff viewer revision
        page
        """
        request = RequestFactory().request()
        request.resolver_match = Mock()
        request.resolver_match.url_name = 'view-diff-revision'

        review_request = self.create_review_request()

        self.assertTrue(self.action.should_render({
            'request': request,
            'review_request': review_request,
        }))

    @add_fixtures(['test_scmtools'])
    def test_should_render_on_review_request_with_repository(self):
        """Testing DownloadDiffAction.should_render on review request page
        with repository
        """
        request = RequestFactory().request()
        request.resolver_match = Mock()
        request.resolver_match.url_name = 'review-request-detail'

        review_request = self.create_review_request(create_repository=True)

        self.assertTrue(self.action.should_render({
            'request': request,
            'review_request': review_request,
        }))

    @add_fixtures(['test_scmtools'])
    def test_should_render_on_review_request_without_repository(self):
        """Testing DownloadDiffAction.should_render on review request page
        without repository
        """
        request = RequestFactory().request()
        request.resolver_match = Mock()
        request.resolver_match.url_name = 'review-request-detail'

        review_request = self.create_review_request()

        self.assertFalse(self.action.should_render({
            'request': request,
            'review_request': review_request,
        }))


class EditReviewActionTests(ActionsTestCase):
    """Unit tests for EditReviewAction."""

    fixtures = ['test_users']

    def setUp(self):
        super(EditReviewActionTests, self).setUp()

        self.action = EditReviewAction()

    def test_should_render_with_authenticated(self):
        """Testing EditReviewAction.should_render with authenticated user"""
        request = RequestFactory().request()
        request.user = User.objects.get(username='doc')

        self.assertTrue(self.action.should_render({'request': request}))

    def test_should_render_with_anonymous(self):
        """Testing EditReviewAction.should_render with authenticated user"""
        request = RequestFactory().request()
        request.user = AnonymousUser()

        self.assertFalse(self.action.should_render({'request': request}))


class ShipItActionTests(ActionsTestCase):
    """Unit tests for ShipItAction."""

    fixtures = ['test_users']

    def setUp(self):
        super(ShipItActionTests, self).setUp()

        self.action = ShipItAction()

    def test_should_render_with_authenticated(self):
        """Testing ShipItAction.should_render with authenticated user"""
        request = RequestFactory().request()
        request.user = User.objects.get(username='doc')

        self.assertTrue(self.action.should_render({'request': request}))

    def test_should_render_with_anonymous(self):
        """Testing ShipItAction.should_render with authenticated user"""
        request = RequestFactory().request()
        request.user = AnonymousUser()

        self.assertFalse(self.action.should_render({'request': request}))


class SubmitActionTests(ActionsTestCase):
    """Unit tests for SubmitAction."""

    fixtures = ['test_users']

    def setUp(self):
        super(SubmitActionTests, self).setUp()

        self.action = SubmitAction()

    def test_should_render_with_published(self):
        """Testing SubmitAction.should_render with published review request"""
        self.assertTrue(self.action.should_render({
            'review_request': self.create_review_request(public=True),
        }))

    def test_should_render_with_unpublished(self):
        """Testing SubmitAction.should_render with unpublished review request
        """
        self.assertFalse(self.action.should_render({
            'review_request': self.create_review_request(public=False),
        }))


class UpdateMenuActionTests(ActionsTestCase):
    """Unit tests for UpdateMenuAction."""

    fixtures = ['test_users']

    def setUp(self):
        super(UpdateMenuActionTests, self).setUp()

        self.action = UpdateMenuAction()

    def test_should_render_for_owner(self):
        """Testing UpdateMenuAction.should_render for owner of review request
        """
        review_request = self.create_review_request(publish=True)

        request = RequestFactory().request()
        request.user = review_request.submitter

        self.assertTrue(self.action.should_render({
            'review_request': review_request,
            'request': request,
            'perms': {
                'reviews': {
                    'can_edit_reviewrequest': False,
                },
            },
        }))

    def test_should_render_for_user(self):
        """Testing UpdateMenuAction.should_render for normal user"""
        review_request = self.create_review_request(publish=True)

        request = RequestFactory().request()
        request.user = User.objects.create_user(username='test-user',
                                                email='user@example.com')

        self.assertFalse(self.action.should_render({
            'review_request': review_request,
            'request': request,
            'perms': {
                'reviews': {
                    'can_edit_reviewrequest': False,
                },
            },
        }))

    def test_should_render_user_with_can_edit_reviewrequest(self):
        """Testing UpdateMenuAction.should_render for user with
        can_edit_reviewrequest permission
        """
        review_request = self.create_review_request(publish=True)

        request = RequestFactory().request()
        request.user = User.objects.create_user(username='test-user',
                                                email='user@example.com')

        self.assertTrue(self.action.should_render({
            'review_request': review_request,
            'request': request,
            'perms': {
                'reviews': {
                    'can_edit_reviewrequest': True,
                },
            },
        }))

    def test_should_render_with_discarded(self):
        """Testing UpdateMenuAction.should_render with discarded review request
        """
        review_request = \
            self.create_review_request(status=ReviewRequest.DISCARDED)

        request = RequestFactory().request()
        request.user = review_request.submitter

        self.assertFalse(self.action.should_render({
            'review_request': review_request,
            'request': request,
            'perms': {
                'reviews': {
                    'can_edit_reviewrequest': False,
                },
            },
        }))

    def test_should_render_with_submitted(self):
        """Testing UpdateMenuAction.should_render with submitted review request
        """
        review_request = \
            self.create_review_request(status=ReviewRequest.SUBMITTED)

        request = RequestFactory().request()
        request.user = review_request.submitter

        self.assertFalse(self.action.should_render({
            'review_request': review_request,
            'request': request,
            'perms': {
                'reviews': {
                    'can_edit_reviewrequest': False,
                },
            },
        }))


class UploadDiffActionTests(ActionsTestCase):
    """Unit tests for UploadDiffAction."""

    fixtures = ['test_users']

    def setUp(self):
        super(UploadDiffActionTests, self).setUp()

        self.action = UploadDiffAction()

    def test_get_label_with_no_diffs(self):
        """Testing UploadDiffAction.get_label with no diffs"""
        review_request = self.create_review_request()

        request = RequestFactory().request()
        request.user = review_request.submitter

        self.assertEqual(
            self.action.get_label({
                'review_request': review_request,
                'request': request,
            }),
            'Upload Diff')

    @add_fixtures(['test_scmtools'])
    def test_get_label_with_diffs(self):
        """Testing UploadDiffAction.get_label with diffs"""
        review_request = self.create_review_request(create_repository=True)
        self.create_diffset(review_request)

        request = RequestFactory().request()
        request.user = review_request.submitter

        self.assertEqual(
            self.action.get_label({
                'review_request': review_request,
                'request': request,
            }),
            'Update Diff')

    @add_fixtures(['test_scmtools'])
    def test_should_render_with_repository(self):
        """Testing UploadDiffAction.should_render with repository"""
        review_request = self.create_review_request(create_repository=True)

        self.assertTrue(self.action.should_render({
            'review_request': review_request,
        }))

    def test_should_render_without_repository(self):
        """Testing UploadDiffAction.should_render without repository"""
        review_request = self.create_review_request()

        self.assertFalse(self.action.should_render({
            'review_request': review_request,
        }))



from __future__ import unicode_literals

from django.core.urlresolvers import NoReverseMatch, reverse


def local_site_reverse(viewname, request=None, local_site_name=None,
                       local_site=None, args=None, kwargs=None,
                       *func_args, **func_kwargs):
    """Reverse a URL name and return a working URL.

    This works much like Django's :py:func:`~django.core.urlresolvers.reverse`,
    but handles returning a LocalSite version of a URL when invoked with one of
    the following:

    * A ``request`` argument, representing an HTTP request to a URL within a
      LocalSite.
    * A ``local_site_name`` argument, indicating the name of the local site.
    * A ``local_site`` argument.

    Args:
        viewname (unicode):
            The name of the view to generate a URL for.

        request (django.http.HttpRequest, optional):
            The current HTTP request. The current local site can be extracted
            from this.

        local_site_name (unicode, optional):
            The name of the local site.

        local_site (reviewboard.site.models.LocalSite, optional):
            The local site.

        args (list, optional):
            Positional arguments to use for reversing in
            :py:func:`~django.core.urlresolvers.reverse`.

        kwargs (dict, optional):
            Keyword arguments to use for reversing in
             :py:func:`~django.core.urlresolvers.reverse`.

        func_args (tuple, optional):
            Additional positional arguments to pass to
            :py:func:`~django.core.urlresolvers.reverse`.

        func_kwargs (dict, optional):
            Additional keyword arguments to pass to
            :py:func:`~django.core.urlresolvers.reverse`.

    Returns:
        unicode:
        The reversed URL.

    Raises:
        django.core.urlresolvers.NoReverseMatch:
            Raised when there is no URL matching the view and arguments.
    """
    assert not (local_site_name and local_site)

    if request or local_site_name or local_site:
        if local_site:
            local_site_name = local_site.name
        elif request and not local_site_name:
            local_site_name = getattr(request, '_local_site_name', None)

        if local_site_name:
            if args:
                new_args = [local_site_name] + list(args)
                new_kwargs = kwargs
            else:
                new_args = args
                new_kwargs = {
                    'local_site_name': local_site_name,
                }

                if kwargs:
                    new_kwargs.update(kwargs)

            try:
                return reverse(viewname, args=new_args, kwargs=new_kwargs,
                               *func_args, **func_kwargs)
            except NoReverseMatch:
                # We'll try it again without those arguments.
                pass

    return reverse(viewname, args=args, kwargs=kwargs,
                   *func_args, **func_kwargs)



"""Unit tests for ReviewRequestPageEntryRegistry."""

from __future__ import unicode_literals

from djblets.registries.errors import AlreadyRegisteredError

from reviewboard.reviews.detail import (BaseReviewRequestPageEntry,
                                        ReviewRequestPageEntryRegistry)
from reviewboard.testing import TestCase


class DummyEntry(BaseReviewRequestPageEntry):
    entry_type_id = 'dummy'


class ReviewRequestPageEntryRegistryTests(TestCase):
    """Unit tests for ReviewRequestPageEntryRegistry."""

    def setUp(self):
        super(ReviewRequestPageEntryRegistryTests, self).setUp()

        self.registry = ReviewRequestPageEntryRegistry()

    def test_register(self):
        """Testing ReviewRequestPageEntryRegistry.register"""
        self.registry.register(DummyEntry)
        self.assertIn(DummyEntry, self.registry)

    def test_register_with_entry_already_registered(self):
        """Testing ReviewRequestPageEntryRegistry.register with already
        registered entry
        """
        self.registry.register(DummyEntry)

        message = 'This review request page entry is already registered.'

        with self.assertRaisesMessage(AlreadyRegisteredError, message):
            self.registry.register(DummyEntry)

    def test_register_with_id_already_registered(self):
        """Testing ReviewRequestPageEntryRegistry.register with already
        registered entry_type_id
        """
        class DummyEntry2(DummyEntry):
            pass

        self.registry.register(DummyEntry)

        message = (
            'A review request page entry with the entry_type_id "dummy" is '
            'already registered by another entry (<class \'reviewboard.'
            'reviews.tests.test_review_request_page_entry_registry.'
            'DummyEntry\'>).'
        )

        with self.assertRaisesMessage(AlreadyRegisteredError, message):
            self.registry.register(DummyEntry2)

    def test_unregister(self):
        """Testing ReviewRequestPageEntryRegistry.unregister"""
        self.registry.register(DummyEntry)
        self.registry.unregister(DummyEntry)

        self.assertNotIn(DummyEntry, self.registry)

    def test_get_entry(self):
        """Testing ReviewRequestPageEntryRegistry.get_entry"""
        self.registry.register(DummyEntry)
        self.assertEqual(self.registry.get_entry('dummy'), DummyEntry)

    def test_get_entry_with_invalid_id(self):
        """Testing ReviewRequestPageEntryRegistry.get_entry with invalid entry
        ID
        """
        self.assertIsNone(self.registry.get_entry('dummy'))



"""Unit tests for the GitHub hosting service."""

from __future__ import unicode_literals

import hashlib
import hmac
import uuid

from django.core.exceptions import ObjectDoesNotExist
from djblets.testing.decorators import add_fixtures

from reviewboard.scmtools.core import Branch, Commit
from reviewboard.hostingsvcs.errors import RepositoryError
from reviewboard.hostingsvcs.repository import RemoteRepository
from reviewboard.hostingsvcs.testing import HostingServiceTestCase
from reviewboard.reviews.models import ReviewRequest
from reviewboard.scmtools.errors import SCMError
from reviewboard.site.models import LocalSite
from reviewboard.site.urlresolvers import local_site_reverse


class GitHubTestCase(HostingServiceTestCase):
    """Base class for GitHub test suites."""

    service_name = 'github'

    default_account_data = {
        'authorization': {
            'token': 'abc123',
        },
    }

    default_repository_extra_data = {
        'repository_plan': 'public',
        'github_public_repo_name': 'myrepo',
    }


class GitHubTests(GitHubTestCase):
    """Unit tests for the GitHub hosting service."""

    def test_service_support(self):
        """Testing GitHub service support capabilities"""
        self.assertTrue(self.service_class.supports_bug_trackers)
        self.assertTrue(self.service_class.supports_repositories)
        self.assertFalse(self.service_class.supports_ssh_key_association)

    def test_get_repository_fields_with_public_plan(self):
        """Testing GitHub.get_repository_fields with the public plan"""
        self.assertEqual(
            self.get_repository_fields(
                'Git',
                plan='public',
                fields={
                    'github_public_repo_name': 'myrepo',
                }
            ),
            {
                'path': 'git://github.com/myuser/myrepo.git',
                'mirror_path': 'git@github.com:myuser/myrepo.git',
            })

    def test_get_repository_fields_with_public_org_plan(self):
        """Testing GitHub.get_repository_fields with the public-org plan"""
        self.assertEqual(
            self.get_repository_fields(
                'Git',
                plan='public-org',
                fields={
                    'github_public_org_repo_name': 'myrepo',
                    'github_public_org_name': 'myorg',
                }
            ),
            {
                'path': 'git://github.com/myorg/myrepo.git',
                'mirror_path': 'git@github.com:myorg/myrepo.git',
            })

    def test_get_repository_fields_with_private_plan(self):
        """Testing GitHub.get_repository_fields with the private plan"""
        self.assertEqual(
            self.get_repository_fields(
                'Git',
                plan='private',
                fields={
                    'github_private_repo_name': 'myrepo',
                }
            ),
            {
                'path': 'git@github.com:myuser/myrepo.git',
                'mirror_path': '',
            })

    def test_get_repository_fields_with_private_org_plan(self):
        """Testing GitHub.get_repository_fields with the private-org plan"""
        self.assertEqual(
            self.get_repository_fields(
                'Git',
                plan='private-org',
                fields={
                    'github_private_org_repo_name': 'myrepo',
                    'github_private_org_name': 'myorg',
                }
            ),
            {
                'path': 'git@github.com:myorg/myrepo.git',
                'mirror_path': '',
            })

    def test_get_repo_api_url_with_public_plan(self):
        """Testing GitHub._get_repo_api_url with the public plan"""
        url = self._get_repo_api_url('public', {
            'github_public_repo_name': 'testrepo',
        })
        self.assertEqual(url, 'https://api.github.com/repos/myuser/testrepo')

    def test_get_repo_api_url_with_public_org_plan(self):
        """Testing GitHub._get_repo_api_url with the public-org plan"""
        url = self._get_repo_api_url('public-org', {
            'github_public_org_name': 'myorg',
            'github_public_org_repo_name': 'testrepo',
        })
        self.assertEqual(url, 'https://api.github.com/repos/myorg/testrepo')

    def test_get_repo_api_url_with_private_plan(self):
        """Testing GitHub._get_repo_api_url with the private plan"""
        url = self._get_repo_api_url('private', {
            'github_private_repo_name': 'testrepo',
        })
        self.assertEqual(url, 'https://api.github.com/repos/myuser/testrepo')

    def test_get_repo_api_url_with_private_org_plan(self):
        """Testing GitHub._get_repo_api_url with the private-org plan"""
        url = self._get_repo_api_url('private-org', {
            'github_private_org_name': 'myorg',
            'github_private_org_repo_name': 'testrepo',
        })
        self.assertEqual(url, 'https://api.github.com/repos/myorg/testrepo')

    def test_get_bug_tracker_field_with_public_plan(self):
        """Testing GitHub.get_bug_tracker_field with the public plan"""
        self.assertTrue(
            self.service_class.get_bug_tracker_requires_username('public'))
        self.assertEqual(
            self.service_class.get_bug_tracker_field('public', {
                'github_public_repo_name': 'myrepo',
                'hosting_account_username': 'myuser',
            }),
            'http://github.com/myuser/myrepo/issues#issue/%s')

    def test_get_bug_tracker_field_with_public_org_plan(self):
        """Testing GitHub.get_bug_tracker_field with the public-org plan"""
        self.assertFalse(
            self.service_class.get_bug_tracker_requires_username('public-org'))
        self.assertEqual(
            self.service_class.get_bug_tracker_field('public-org', {
                'github_public_org_name': 'myorg',
                'github_public_org_repo_name': 'myrepo',
            }),
            'http://github.com/myorg/myrepo/issues#issue/%s')

    def test_get_bug_tracker_field_with_private_plan(self):
        """Testing GitHub.get_bug_tracker_field with the private plan"""
        self.assertTrue(
            self.service_class.get_bug_tracker_requires_username('private'))
        self.assertEqual(
            self.service_class.get_bug_tracker_field('private', {
                'github_private_repo_name': 'myrepo',
                'hosting_account_username': 'myuser',
            }),
            'http://github.com/myuser/myrepo/issues#issue/%s')

    def test_get_bug_tracker_field_with_private_org_plan(self):
        """Testing GitHub.get_bug_tracker_field with the private-org plan"""
        self.assertFalse(self.service_class.get_bug_tracker_requires_username(
            'private-org'))
        self.assertEqual(
            self.service_class.get_bug_tracker_field('private-org', {
                'github_private_org_name': 'myorg',
                'github_private_org_repo_name': 'myrepo',
            }),
            'http://github.com/myorg/myrepo/issues#issue/%s')

    def test_check_repository_public(self):
        """Testing GitHub.check_repository with public repository"""
        self._test_check_repository(plan='public',
                                    github_public_repo_name='myrepo')

    def test_check_repository_private(self):
        """Testing GitHub.check_repository with private repository"""
        self._test_check_repository(plan='private',
                                    github_private_repo_name='myrepo')

    def test_check_repository_public_org(self):
        """Testing GitHub.check_repository with public org repository"""
        self._test_check_repository(plan='public-org',
                                    github_public_org_name='myorg',
                                    github_public_org_repo_name='myrepo',
                                    expected_owner='myorg')

    def test_check_repository_private_org(self):
        """Testing GitHub.check_repository with private org repository"""
        self._test_check_repository(plan='private-org',
                                    github_private_org_name='myorg',
                                    github_private_org_repo_name='myrepo',
                                    expected_owner='myorg')

    def test_check_repository_public_not_found(self):
        """Testing GitHub.check_repository with not found error and public
        repository
        """
        self._test_check_repository_error(
            plan='public',
            github_public_repo_name='myrepo',
            http_status=404,
            payload=b'{"message": "Not Found"}',
            expected_url='https://api.github.com/repos/myuser/myrepo',
            expected_error='A repository with this name was not found, '
                           'or your user may not own it.')

    def test_check_repository_private_not_found(self):
        """Testing GitHub.check_repository with not found error and private
        repository
        """
        self._test_check_repository_error(
            plan='private',
            github_private_repo_name='myrepo',
            http_status=404,
            payload=b'{"message": "Not Found"}',
            expected_url='https://api.github.com/repos/myuser/myrepo',
            expected_error='A repository with this name was not found, '
                           'or your user may not own it.')

    def test_check_repository_public_org_not_found(self):
        """Testing GitHub.check_repository with not found error and
        public organization repository
        """
        self._test_check_repository_error(
            plan='public-org',
            github_public_org_name='myorg',
            github_public_org_repo_name='myrepo',
            http_status=404,
            payload=b'{"message": "Not Found"}',
            expected_url='https://api.github.com/repos/myorg/myrepo',
            expected_error='A repository with this organization or name '
                           'was not found.')

    def test_check_repository_private_org_not_found(self):
        """Testing GitHub.check_repository with not found error and
        private organization repository
        """
        self._test_check_repository_error(
            plan='private-org',
            github_private_org_name='myorg',
            github_private_org_repo_name='myrepo',
            http_status=404,
            payload=b'{"message": "Not Found"}',
            expected_url='https://api.github.com/repos/myorg/myrepo',
            expected_error='A repository with this organization or name '
                           'was not found, or your user may not have access '
                           'to it.')

    def test_check_repository_public_plan_private_repo(self):
        """Testing GitHub.check_repository with public plan and
        private repository
        """
        self._test_check_repository_error(
            plan='public',
            github_public_repo_name='myrepo',
            http_status=200,
            payload=b'{"private": true}',
            expected_url='https://api.github.com/repos/myuser/myrepo',
            expected_error='This is a private repository, but you have '
                           'selected a public plan.')

    def test_check_repository_private_plan_public_repo(self):
        """Testing GitHub.check_repository with private plan and
        public repository
        """
        self._test_check_repository_error(
            plan='private',
            github_private_repo_name='myrepo',
            http_status=200,
            payload=b'{"private": false}',
            expected_url='https://api.github.com/repos/myuser/myrepo',
            expected_error='This is a public repository, but you have '
                           'selected a private plan.')

    def test_check_repository_public_org_plan_private_repo(self):
        """Testing GitHub.check_repository with public organization plan and
        private repository
        """
        self._test_check_repository_error(
            plan='public-org',
            github_public_org_name='myorg',
            github_public_org_repo_name='myrepo',
            http_status=200,
            payload=b'{"private": true}',
            expected_url='https://api.github.com/repos/myorg/myrepo',
            expected_error='This is a private repository, but you have '
                           'selected a public plan.')

    def test_check_repository_private_org_plan_public_repo(self):
        """Testing GitHub.check_repository with private organization plan and
        public repository
        """
        self._test_check_repository_error(
            plan='private-org',
            github_private_org_name='myorg',
            github_private_org_repo_name='myrepo',
            http_status=200,
            payload=b'{"private": false}',
            expected_url='https://api.github.com/repos/myorg/myrepo',
            expected_error='This is a public repository, but you have '
                           'selected a private plan.')

    def test_authorization(self):
        """Testing GitHub.authorize"""
        payload = self.dump_json({
            'id': 1,
            'url': 'https://api.github.com/authorizations/1',
            'scopes': ['user', 'repo'],
            'token': 'abc123',
            'note': '',
            'note_url': '',
            'updated_at': '2012-05-04T03:30:00Z',
            'created_at': '2012-05-04T03:30:00Z',
        })

        hosting_account = self.create_hosting_account(data={})
        self.assertFalse(hosting_account.is_authorized)

        self.spy_on(
            uuid.uuid4,
            call_fake=lambda: uuid.UUID('2a707f8c6fc14dd590e545ebe1e9b2f6'))

        with self.setup_http_test(payload=payload,
                                  hosting_account=hosting_account,
                                  expected_http_calls=1) as ctx:
            with self.settings(GITHUB_CLIENT_ID=None,
                               GITHUB_CLIENT_SECRET=None):
                ctx.service.authorize(username='myuser',
                                      password='mypass')

        self.assertTrue(hosting_account.is_authorized)

        ctx.assertHTTPCall(
            0,
            url='https://api.github.com/authorizations',
            method='POST',
            username='myuser',
            password='mypass',
            body=(
                '{'
                '"note": "Access for Review Board (example.com/ - 2a707f8)", '
                '"note_url": "http://example.com/", '
                '"scopes": ["user", "repo"]'
                '}'
            ),
            headers={
                'Content-Length': '123',
            })

    def test_authorization_with_client_info(self):
        """Testing GitHub.authorize with registered client ID/secret"""
        payload = self.dump_json({
            'id': 1,
            'url': 'https://api.github.com/authorizations/1',
            'scopes': ['user', 'repo'],
            'token': 'abc123',
            'note': '',
            'note_url': '',
            'updated_at': '2012-05-04T03:30:00Z',
            'created_at': '2012-05-04T03:30:00Z',
        })

        hosting_account = self.create_hosting_account(data={})
        self.assertFalse(hosting_account.is_authorized)

        self.spy_on(
            uuid.uuid4,
            call_fake=lambda: uuid.UUID('2a707f8c6fc14dd590e545ebe1e9b2f6'))

        with self.setup_http_test(payload=payload,
                                  hosting_account=hosting_account,
                                  expected_http_calls=1) as ctx:
            with self.settings(GITHUB_CLIENT_ID='abc123',
                               GITHUB_CLIENT_SECRET='def456'):
                ctx.service.authorize(username='myuser',
                                      password='mypass')

        self.assertTrue(hosting_account.is_authorized)

        ctx.assertHTTPCall(
            0,
            url='https://api.github.com/authorizations',
            method='POST',
            username='myuser',
            password='mypass',
            body=(
                '{'
                '"client_id": "abc123", '
                '"client_secret": "def456", '
                '"note": "Access for Review Board (example.com/ - 2a707f8)", '
                '"note_url": "http://example.com/", '
                '"scopes": ["user", "repo"]'
                '}'
            ),
            headers={
                'Content-Length': '173',
            })

    def test_get_branches(self):
        """Testing GitHub.get_branches"""
        payload = self.dump_json([
            {
                'ref': 'refs/heads/master',
                'object': {
                    'sha': '859d4e148ce3ce60bbda6622cdbe5c2c2f8d9817',
                }
            },
            {
                'ref': 'refs/heads/release-1.7.x',
                'object': {
                    'sha': '92463764015ef463b4b6d1a1825fee7aeec8cb15',
                }
            },
            {
                'ref': 'refs/heads/some-component/fix',
                'object': {
                    'sha': '764015ef492c8cb1546363b45fee7ab6d1a182ee',
                }
            },
            {
                'ref': 'refs/tags/release-1.7.11',
                'object': {
                    'sha': 'f5a35f1d8a8dcefb336a8e3211334f1f50ea7792',
                }
            },
        ])

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=1) as ctx:
            repository = ctx.create_repository()
            branches = ctx.service.get_branches(repository)

        ctx.assertHTTPCall(
            0,
            url=('https://api.github.com/repos/myuser/myrepo/git/refs/heads'
                 '?access_token=abc123'),
            username=None,
            password=None)

        self.assertEqual(
            branches,
            [
                Branch(id='master',
                       commit='859d4e148ce3ce60bbda6622cdbe5c2c2f8d9817',
                       default=True),
                Branch(id='release-1.7.x',
                       commit='92463764015ef463b4b6d1a1825fee7aeec8cb15',
                       default=False),
                Branch(id='some-component/fix',
                       commit='764015ef492c8cb1546363b45fee7ab6d1a182ee',
                       default=False),
            ])

    def test_get_commits(self):
        """Testing GitHub.get_commits"""
        payload = self.dump_json([
            {
                'commit': {
                    'author': {'name': 'Christian Hammond'},
                    'committer': {'date': '2013-06-25T23:31:22Z'},
                    'message': 'Fixed the bug number for the '
                               'blacktriangledown bug.',
                },
                'sha': '859d4e148ce3ce60bbda6622cdbe5c2c2f8d9817',
                'parents': [
                    {'sha': '92463764015ef463b4b6d1a1825fee7aeec8cb15'}
                ],
            },
            {
                'commit': {
                    'author': {'name': 'Christian Hammond'},
                    'committer': {'date': '2013-06-25T23:30:59Z'},
                    'message': "Merge branch 'release-1.7.x'",
                },
                'sha': '92463764015ef463b4b6d1a1825fee7aeec8cb15',
                'parents': [
                    {'sha': 'f5a35f1d8a8dcefb336a8e3211334f1f50ea7792'},
                    {'sha': '6c5f3465da5ed03dca8128bb3dd03121bd2cddb2'},
                ],
            },
            {
                'commit': {
                    'author': {'name': 'David Trowbridge'},
                    'committer': {'date': '2013-06-25T22:41:09Z'},
                    'message': 'Add DIFF_PARSE_ERROR to the '
                               'ValidateDiffResource.create error list.',
                },
                'sha': 'f5a35f1d8a8dcefb336a8e3211334f1f50ea7792',
                'parents': [],
            }
        ])

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=1) as ctx:
            repository = ctx.create_repository()
            commits = ctx.service.get_commits(
                repository=repository,
                start='859d4e148ce3ce60bbda6622cdbe5c2c2f8d9817')

        ctx.assertHTTPCall(
            0,
            url=('https://api.github.com/repos/myuser/myrepo/commits'
                 '?access_token=abc123'
                 '&sha=859d4e148ce3ce60bbda6622cdbe5c2c2f8d9817'),
            username=None,
            password=None)

        self.assertEqual(
            commits,
            [
                Commit(author_name='Christian Hammond',
                       date='2013-06-25T23:31:22Z',
                       id='859d4e148ce3ce60bbda6622cdbe5c2c2f8d9817',
                       message=('Fixed the bug number for the '
                                'blacktriangledown bug.'),
                       parent='92463764015ef463b4b6d1a1825fee7aeec8cb15'),
                Commit(author_name='Christian Hammond',
                       date='2013-06-25T23:30:59Z',
                       id='92463764015ef463b4b6d1a1825fee7aeec8cb15',
                       message="Merge branch 'release-1.7.x'",
                       parent='f5a35f1d8a8dcefb336a8e3211334f1f50ea7792'),
                Commit(author_name='David Trowbridge',
                       date='2013-06-25T22:41:09Z',
                       id='f5a35f1d8a8dcefb336a8e3211334f1f50ea7792',
                       message=('Add DIFF_PARSE_ERROR to the '
                                'ValidateDiffResource.create error list.'),
                       parent=''),
            ])

        for commit in commits:
            self.assertIsNone(commit.diff)

    def test_get_change(self):
        """Testing GitHub.get_change"""
        commit_sha = '1c44b461cebe5874a857c51a4a13a849a4d1e52d'
        parent_sha = '44568f7d33647d286691517e6325fea5c7a21d5e'
        tree_sha = '56e25e58380daf9b4dfe35677ae6043fe1743922'

        paths = {
            '/repos/myuser/myrepo/commits': {
                'payload': self.dump_json([
                    {
                        'commit': {
                            'author': {'name': 'David Trowbridge'},
                            'committer': {'date': '2013-06-25T23:31:22Z'},
                            'message': 'Move .clearfix to defs.less',
                        },
                        'sha': commit_sha,
                        'parents': [{'sha': parent_sha}],
                    },
                ])
            },
            '/repos/myuser/myrepo/compare/%s...%s' % (parent_sha,
                                                      commit_sha): {
                'payload': self.dump_json({
                    'base_commit': {
                        'commit': {
                            'tree': {'sha': tree_sha},
                        },
                    },
                    'files': [
                        {
                            'sha': '4344b3ad41b171ea606e88e9665c34cca602affb',
                            'filename': 'reviewboard/static/rb/css/defs.less',
                            'status': 'modified',
                            'patch': (
                                '@@ -182,4 +182,6 @@\n'
                                ' }\n'
                                ' \n'
                                '+.foo {\n'
                                '+}\n'
                                ' \n'
                                ' table {'
                            ),
                        },
                        {
                            'sha': '8e3129277b018b169cb8d13771433fbcd165a17c',
                            'filename': ('reviewboard/static/rb/css/'
                                         'reviews.less'),
                            'status': 'modified',
                            'patch': (
                                '@@ -1311,6 +1311,4 @@\n'
                                ' }\n'
                                ' \n'
                                '-.bar {\n'
                                '-}\n'
                                ' \n'
                                ' h1 {'
                            ),
                        },
                        {
                            'sha': '17ba0791499db908433b80f37c5fbc89b870084b',
                            'filename': 'new_filename',
                            'previous_filename': 'old_filename',
                            'status': 'renamed',
                            'patch': (
                                '@@ -1,1 +1,1 @@\n'
                                '- foo\n'
                                '+ bar'
                            ),
                        },
                    ],
                }),
            },
            '/repos/myuser/myrepo/git/trees/%s' % tree_sha: {
                'payload': self.dump_json({
                    'tree': [
                        {
                            'path': 'reviewboard/static/rb/css/defs.less',
                            'sha': '830a40c3197223c6a0abb3355ea48891a1857bfd',
                        },
                        {
                            'path': 'reviewboard/static/rb/css/reviews.less',
                            'sha': '535cd2c4211038d1bb8ab6beaed504e0db9d7e62',
                        },
                        {
                            'path': 'old_filename',
                            'sha': '356a192b7913b04c54574d18c28d46e6395428ab',
                        }
                    ],
                }),
            },
        }

        with self.setup_http_test(self.make_handler_for_paths(paths),
                                  expected_http_calls=3) as ctx:
            repository = ctx.create_repository()
            change = ctx.service.get_change(repository=repository,
                                            revision=commit_sha)

        ctx.assertHTTPCall(
            0,
            url=('https://api.github.com/repos/myuser/myrepo/commits'
                 '?access_token=abc123'
                 '&sha=1c44b461cebe5874a857c51a4a13a849a4d1e52d'),
            username=None,
            password=None)

        ctx.assertHTTPCall(
            1,
            url=('https://api.github.com/repos/myuser/myrepo/compare/'
                 '44568f7d33647d286691517e6325fea5c7a21d5e...'
                 '1c44b461cebe5874a857c51a4a13a849a4d1e52d'
                 '?access_token=abc123'),
            username=None,
            password=None)

        ctx.assertHTTPCall(
            2,
            url=('https://api.github.com/repos/myuser/myrepo/git/trees/'
                 '56e25e58380daf9b4dfe35677ae6043fe1743922'
                 '?access_token=abc123&recursive=1'),
            username=None,
            password=None)

        self.assertEqual(
            change,
            Commit(author_name='David Trowbridge',
                   date='2013-06-25T23:31:22Z',
                   id='1c44b461cebe5874a857c51a4a13a849a4d1e52d',
                   message='Move .clearfix to defs.less',
                   parent='44568f7d33647d286691517e6325fea5c7a21d5e'))
        self.assertEqual(
            change.diff,
            b'diff --git a/reviewboard/static/rb/css/defs.less'
            b' b/reviewboard/static/rb/css/defs.less\n'
            b'index 830a40c3197223c6a0abb3355ea48891a1857bfd'
            b'..4344b3ad41b171ea606e88e9665c34cca602affb 100644\n'
            b'--- a/reviewboard/static/rb/css/defs.less\n'
            b'+++ b/reviewboard/static/rb/css/defs.less\n'
            b'@@ -182,4 +182,6 @@\n'
            b' }\n'
            b' \n'
            b'+.foo {\n'
            b'+}\n'
            b' \n'
            b' table {\n'
            b'diff --git a/reviewboard/static/rb/css/reviews.less'
            b' b/reviewboard/static/rb/css/reviews.less\n'
            b'index 535cd2c4211038d1bb8ab6beaed504e0db9d7e62'
            b'..8e3129277b018b169cb8d13771433fbcd165a17c 100644\n'
            b'--- a/reviewboard/static/rb/css/reviews.less\n'
            b'+++ b/reviewboard/static/rb/css/reviews.less\n'
            b'@@ -1311,6 +1311,4 @@\n'
            b' }\n'
            b' \n'
            b'-.bar {\n'
            b'-}\n'
            b' \n'
            b' h1 {\n'
            b'diff --git a/new_filename b/new_filename\n'
            b'rename from old_filename\n'
            b'rename to new_filename\n'
            b'index 356a192b7913b04c54574d18c28d46e6395428ab'
            b'..17ba0791499db908433b80f37c5fbc89b870084b\n'
            b'--- a/old_filename\n'
            b'+++ b/new_filename\n'
            b'@@ -1,1 +1,1 @@\n'
            b'- foo\n'
            b'+ bar\n')

    def test_get_change_with_not_found(self):
        """Testing GitHub.get_change with commit not found"""
        with self.setup_http_test(status_code=404,
                                  payload=b'{"message": "Not Found"}',
                                  expected_http_calls=1) as ctx:
            with self.assertRaisesMessage(SCMError, 'Not Found'):
                repository = ctx.create_repository()
                ctx.service.get_change(
                    repository=repository,
                    revision='1c44b461cebe5874a857c51a4a13a849a4d1e52d')

        ctx.assertHTTPCall(
            0,
            url=('https://api.github.com/repos/myuser/myrepo/commits'
                 '?access_token=abc123'
                 '&sha=1c44b461cebe5874a857c51a4a13a849a4d1e52d'),
            username=None,
            password=None)

    def test_get_remote_repositories_with_owner(self):
        """Testing GitHub.get_remote_repositories with requesting
        authenticated user's repositories
        """
        base_url = 'https://api.github.com/user/repos?access_token=abc123'
        paths = {
            '/user/repos?access_token=abc123': {
                'payload': self.dump_json([
                    {
                        'id': 1,
                        'owner': {
                            'login': 'myuser',
                        },
                        'name': 'myrepo',
                        'clone_url': 'myrepo_path',
                        'mirror_url': 'myrepo_mirror',
                        'private': 'false'
                    },
                ]),
                'headers': {
                    b'Link': b'<%s&page=2>; rel="next"' % base_url,
                },
            },
            '/user/repos?access_token=abc123&page=2': {
                'payload': self.dump_json([
                    {
                        'id': 2,
                        'owner': {
                            'login': 'myuser',
                        },
                        'name': 'myrepo2',
                        'clone_url': 'myrepo_path2',
                        'mirror_url': 'myrepo_mirror2',
                        'private': 'true'
                    },
                ]),
                'headers': {
                    b'Link': b'<%s&page=1>; rel="prev"' % base_url,
                },
            },
        }

        # Fetch and check the first page.
        with self.setup_http_test(self.make_handler_for_paths(paths),
                                  expected_http_calls=1) as ctx:
            paginator = ctx.service.get_remote_repositories('myuser')

        ctx.assertHTTPCall(
            0,
            url='https://api.github.com/user/repos?access_token=abc123',
            username=None,
            password=None)

        self.assertEqual(len(paginator.page_data), 1)
        self.assertFalse(paginator.has_prev)
        self.assertTrue(paginator.has_next)
        repo = paginator.page_data[0]

        self.assertIsInstance(repo, RemoteRepository)
        self.assertEqual(repo.id, 'myuser/myrepo')
        self.assertEqual(repo.owner, 'myuser')
        self.assertEqual(repo.name, 'myrepo')
        self.assertEqual(repo.scm_type, 'Git')
        self.assertEqual(repo.path, 'myrepo_path')
        self.assertEqual(repo.mirror_path, 'myrepo_mirror')

        # Fetch and check the second page.
        paginator.next()

        ctx.assertHTTPCall(
            1,
            url='https://api.github.com/user/repos?access_token=abc123&page=2',
            username=None,
            password=None)

        self.assertEqual(len(paginator.page_data), 1)
        self.assertTrue(paginator.has_prev)
        self.assertFalse(paginator.has_next)
        repo = paginator.page_data[0]

        self.assertIsInstance(repo, RemoteRepository)
        self.assertEqual(repo.id, 'myuser/myrepo2')
        self.assertEqual(repo.owner, 'myuser')
        self.assertEqual(repo.name, 'myrepo2')
        self.assertEqual(repo.scm_type, 'Git')
        self.assertEqual(repo.path, 'myrepo_path2')
        self.assertEqual(repo.mirror_path, 'myrepo_mirror2')

    def test_get_remote_repositories_with_other_user(self):
        """Testing GitHub.get_remote_repositories with requesting user's
        repositories
        """
        repos1 = self.dump_json([
            {
                'id': 1,
                'owner': {
                    'login': 'other',
                },
                'name': 'myrepo',
                'clone_url': 'myrepo_path',
                'mirror_url': 'myrepo_mirror',
                'private': 'false'
            }
        ])

        headers = {
            b'Link': (
                b'<https://api.github.com/users/other/repos'
                b'?access_token=abc123&page=2>; rel="next"'
            ),
        }

        with self.setup_http_test(payload=repos1,
                                  headers=headers,
                                  expected_http_calls=1) as ctx:
            paginator = ctx.service.get_remote_repositories('other')

        ctx.assertHTTPCall(
            0,
            url='https://api.github.com/users/other/repos?access_token=abc123',
            username=None,
            password=None)

        self.assertEqual(len(paginator.page_data), 1)
        public_repo = paginator.page_data[0]
        self.assertIsInstance(public_repo, RemoteRepository)
        self.assertEqual(public_repo.id, 'other/myrepo')
        self.assertEqual(public_repo.owner, 'other')
        self.assertEqual(public_repo.name, 'myrepo')
        self.assertEqual(public_repo.scm_type, 'Git')
        self.assertEqual(public_repo.path, 'myrepo_path')
        self.assertEqual(public_repo.mirror_path, 'myrepo_mirror')

    def test_get_remote_repositories_with_org(self):
        """Testing GitHub.get_remote_repositories with requesting
        organization's repositories
        """
        payload = self.dump_json([
            {
                'id': 1,
                'owner': {
                    'login': 'myorg',
                },
                'name': 'myrepo',
                'clone_url': 'myrepo_path',
                'mirror_url': 'myrepo_mirror',
                'private': 'false'
            },
            {
                'id': 2,
                'owner': {
                    'login': 'myuser',
                },
                'name': 'myrepo2',
                'clone_url': 'myrepo_path2',
                'mirror_url': 'myrepo_mirror2',
                'private': 'true'
            }
        ])

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=1) as ctx:
            paginator = ctx.service.get_remote_repositories('myorg',
                                                            'organization')

        ctx.assertHTTPCall(
            0,
            url='https://api.github.com/orgs/myorg/repos?access_token=abc123',
            username=None,
            password=None)

        self.assertEqual(len(paginator.page_data), 2)
        public_repo, private_repo = paginator.page_data

        self.assertIsInstance(public_repo, RemoteRepository)
        self.assertEqual(public_repo.id, 'myorg/myrepo')
        self.assertEqual(public_repo.owner, 'myorg')
        self.assertEqual(public_repo.name, 'myrepo')
        self.assertEqual(public_repo.scm_type, 'Git')
        self.assertEqual(public_repo.path, 'myrepo_path')
        self.assertEqual(public_repo.mirror_path, 'myrepo_mirror')

        self.assertIsInstance(private_repo, RemoteRepository)
        self.assertEqual(private_repo.id, 'myuser/myrepo2')
        self.assertEqual(private_repo.owner, 'myuser')
        self.assertEqual(private_repo.name, 'myrepo2')
        self.assertEqual(private_repo.scm_type, 'Git')
        self.assertEqual(private_repo.path, 'myrepo_path2')
        self.assertEqual(private_repo.mirror_path, 'myrepo_mirror2')

    def test_get_remote_repositories_with_defaults(self):
        """Testing GitHub.get_remote_repositories with default values"""
        with self.setup_http_test(payload=b'{}',
                                  expected_http_calls=1) as ctx:
            ctx.service.get_remote_repositories()

        ctx.assertHTTPCall(
            0,
            url='https://api.github.com/user/repos?access_token=abc123',
            username=None,
            password=None)

    def test_get_remote_repositories_with_filter(self):
        """Testing GitHub.get_remote_repositories with ?filter-type="""
        with self.setup_http_test(payload=b'[]',
                                  expected_http_calls=1) as ctx:
            ctx.service.get_remote_repositories('myuser',
                                                filter_type='private')

        ctx.assertHTTPCall(
            0,
            url=('https://api.github.com/user/repos?access_token=abc123'
                 '&type=private'),
            username=None,
            password=None)

    def test_get_remote_repository(self):
        """Testing GitHub.get_remote_repository"""
        payload = self.dump_json({
            'id': 1,
            'owner': {
                'login': 'myuser',
            },
            'name': 'myrepo',
            'clone_url': 'myrepo_path',
            'mirror_url': 'myrepo_mirror',
            'private': 'false'
        })

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=1) as ctx:
            remote_repository = \
                ctx.service.get_remote_repository('myuser/myrepo')

        ctx.assertHTTPCall(
            0,
            url=('https://api.github.com/repos/myuser/myrepo'
                 '?access_token=abc123'),
            username=None,
            password=None)

        self.assertIsInstance(remote_repository, RemoteRepository)
        self.assertEqual(remote_repository.id, 'myuser/myrepo')
        self.assertEqual(remote_repository.owner, 'myuser')
        self.assertEqual(remote_repository.name, 'myrepo')
        self.assertEqual(remote_repository.scm_type, 'Git')
        self.assertEqual(remote_repository.path, 'myrepo_path')
        self.assertEqual(remote_repository.mirror_path, 'myrepo_mirror')

    def test_get_remote_repository_invalid(self):
        """Testing GitHub.get_remote_repository with invalid repository ID"""
        with self.setup_http_test(status_code=404,
                                  payload=b'{"message": "Not Found"}',
                                  expected_http_calls=1) as ctx:
            with self.assertRaises(ObjectDoesNotExist):
                ctx.service.get_remote_repository('myuser/invalid')

        ctx.assertHTTPCall(
            0,
            url=('https://api.github.com/repos/myuser/invalid'
                 '?access_token=abc123'),
            username=None,
            password=None)

    def _test_check_repository(self, expected_owner='myuser', **kwargs):
        """Test checking for a repository.

        Args:
            expected_owner (unicode):
                The expected owner of the repository.

            **kwargs (dict):
                Keyword arguments to pass to
                :py:meth:`check_repository()
                <reviewboard.hostingsvcs.gitlab.GitLab.check_repository>`.
        """
        with self.setup_http_test(payload=b'{}',
                                  expected_http_calls=1) as ctx:
            ctx.service.check_repository(**kwargs)

        ctx.assertHTTPCall(
            0,
            url=('https://api.github.com/repos/%s/myrepo?access_token=abc123'
                 % expected_owner),
            username=None,
            password=None)

    def _test_check_repository_error(self, http_status, payload, expected_url,
                                     expected_error, **kwargs):
        """Test error conditions when checking for a repository.

        Args:
            http_status (int):
                The HTTP status to simulate returning.

            payload (bytes):
                The payload to return, if ``http_status`` is 200.

            expected_url (unicode):
                The expected URL accessed (minus any query strings).

            expected_error (unicode):
                The expected error message from a raised exception.

            **kwargs (dict):
                Keyword arguments to pass to
                :py:meth:`check_repository()
                <reviewboard.hostingsvcs.gitlab.GitLab.check_repository>`.
        """
        if http_status != 200:
            payload = b'{"message": "not Found"}'

        with self.setup_http_test(status_code=http_status,
                                  payload=payload,
                                  expected_http_calls=1) as ctx:
            with self.assertRaisesMessage(RepositoryError, expected_error):
                ctx.service.check_repository(**kwargs)

        ctx.assertHTTPCall(
            0,
            url='%s?access_token=abc123' % expected_url,
            username=None,
            password=None)

    def _get_repo_api_url(self, plan, fields):
        """Return the base API URL for a repository.

        Args:
            plan (unicode):
                The name of the plan.

            fields (dict):
                Fields containing repository information.

        Returns:
            unicode:
            The API URL for the repository.
        """
        account = self.create_hosting_account()

        repository = self.create_repository(
            hosting_account=account,
            extra_data={
                'repository_plan': plan,
            })

        form = self.get_form(plan, fields)
        form.save(repository)

        return account.service._get_repo_api_url(repository)


class CloseSubmittedHookTests(GitHubTestCase):
    """Unit tests for the GitHub close-submitted webhook."""

    fixtures = ['test_users', 'test_scmtools']

    def test_close_submitted_hook(self):
        """Testing GitHub close_submitted hook with event=push"""
        self._test_post_commit_hook()

    @add_fixtures(['test_site'])
    def test_close_submitted_hook_with_local_site(self):
        """Testing GitHub close_submitted hook with event=push and using a
        Local Site
        """
        self._test_post_commit_hook(
            LocalSite.objects.get(name=self.local_site_name))

    @add_fixtures(['test_site'])
    def test_close_submitted_hook_with_unpublished_review_request(self):
        """Testing GitHub close_submitted hook with event=push and an
        un-published review request
        """
        self._test_post_commit_hook(publish=False)

    def test_close_submitted_hook_ping(self):
        """Testing GitHub close_submitted hook with event=ping"""
        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)

        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'github-hooks-close-submitted',
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'github',
            })

        response = self._post_commit_hook_payload(
            url, review_request, repository.get_or_create_hooks_uuid(),
            event='ping')
        self.assertEqual(response.status_code, 200)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)
        self.assertEqual(review_request.changedescs.count(), 0)

    def test_close_submitted_hook_with_invalid_repo(self):
        """Testing GitHub close_submitted hook with event=push and invalid
        repository
        """
        repository = self.create_repository()

        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'github-hooks-close-submitted',
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'github',
            })

        response = self._post_commit_hook_payload(
            url, review_request, repository.get_or_create_hooks_uuid())
        self.assertEqual(response.status_code, 404)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)
        self.assertEqual(review_request.changedescs.count(), 0)

    def test_close_submitted_hook_with_invalid_site(self):
        """Testing GitHub close_submitted hook with event=push and invalid
        Local Site
        """
        repository = self.create_repository()

        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'github-hooks-close-submitted',
            local_site_name='badsite',
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'github',
            })

        response = self._post_commit_hook_payload(
            url, review_request, repository.get_or_create_hooks_uuid())
        self.assertEqual(response.status_code, 404)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)
        self.assertEqual(review_request.changedescs.count(), 0)

    def test_close_submitted_hook_with_invalid_service_id(self):
        """Testing GitHub close_submitted hook with event=push and invalid
        hosting service ID
        """
        # We'll test against Bitbucket for this test.
        account = self.create_hosting_account()
        account.service_name = 'bitbucket'
        account.save()

        repository = self.create_repository(hosting_account=account)

        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'github-hooks-close-submitted',
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'github',
            })

        response = self._post_commit_hook_payload(
            url, review_request, repository.get_or_create_hooks_uuid())
        self.assertEqual(response.status_code, 404)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)
        self.assertEqual(review_request.changedescs.count(), 0)

    def test_close_submitted_hook_with_invalid_event(self):
        """Testing GitHub close_submitted hook with invalid event"""
        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)
        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'github-hooks-close-submitted',
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'github',
            })

        response = self._post_commit_hook_payload(
            url, review_request, repository.get_or_create_hooks_uuid(),
            event='foo')
        self.assertEqual(response.status_code, 400)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)
        self.assertEqual(review_request.changedescs.count(), 0)

    def test_close_submitted_hook_with_invalid_signature(self):
        """Testing GitHub close_submitted hook with invalid signature"""
        account = self.create_hosting_account()
        repository = self.create_repository(hosting_account=account)
        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'github-hooks-close-submitted',
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'github',
            })

        response = self._post_commit_hook_payload(
            url, review_request, 'bad-secret')
        self.assertEqual(response.status_code, 400)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)
        self.assertEqual(review_request.changedescs.count(), 0)

    def _test_post_commit_hook(self, local_site=None, publish=True):
        """Testing posting to a commit hook.

        This will simulate pushing a commit and posting the resulting webhook
        payload from GitHub to the handler for the hook.

        Args:
            local_site (reviewboard.site.models.LocalSite, optional):
                The Local Site owning the review request.

            publish (bool, optional):
                Whether to test with a published review request.
        """
        account = self.create_hosting_account(local_site=local_site)
        repository = self.create_repository(hosting_account=account,
                                            local_site=local_site)

        review_request = self.create_review_request(repository=repository,
                                                    local_site=local_site,
                                                    publish=publish)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'github-hooks-close-submitted',
            local_site=local_site,
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'github',
            })

        response = self._post_commit_hook_payload(
            url, review_request, repository.get_or_create_hooks_uuid())
        self.assertEqual(response.status_code, 200)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.SUBMITTED)
        self.assertEqual(review_request.changedescs.count(), 1)

        changedesc = review_request.changedescs.get()
        self.assertEqual(changedesc.text, 'Pushed to master (1c44b46)')

    def _post_commit_hook_payload(self, url, review_request, secret,
                                  event='push'):
        """Post a payload for a hook for testing.

        Args:
            url (unicode):
                The URL to post to.

            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request being represented in the payload.

            secret (unicode):
                The HMAC secret for the message.

            event (unicode, optional):
                The webhook event.

        Results:
            django.core.handlers.request.wsgi.WSGIRequest:
            The post request.
        """
        payload = self.dump_json({
            # NOTE: This payload only contains the content we make
            #       use of in the hook.
            'ref': 'refs/heads/master',
            'commits': [
                {
                    'id': '1c44b461cebe5874a857c51a4a13a849a4d1e52d',
                    'message': 'This is my fancy commit\n'
                               '\n'
                               'Reviewed at http://example.com%s'
                               % review_request.get_absolute_url(),
                },
            ]
        })

        m = hmac.new(bytes(secret), payload, hashlib.sha1)

        return self.client.post(
            url,
            payload,
            content_type='application/json',
            HTTP_X_GITHUB_EVENT=event,
            HTTP_X_HUB_SIGNATURE='sha1=%s' % m.hexdigest())



from __future__ import unicode_literals

from django.db import models
from django.utils.translation import ugettext_lazy as _
from djblets.webapi.models import BaseWebAPIToken

from reviewboard.site.models import LocalSite


class WebAPIToken(BaseWebAPIToken):
    """An access token used for authenticating with the API.

    Each token can be used to authenticate the token's owner with the API,
    without requiring a username or password to be provided. Tokens can
    be revoked, and new tokens added.

    Tokens can store policy information, which will later be used for
    restricting access to the API.
    """

    local_site = models.ForeignKey(LocalSite, related_name='webapi_tokens',
                                   blank=True, null=True)

    @classmethod
    def get_root_resource(self):
        from reviewboard.webapi.resources import resources

        return resources.root

    class Meta:
        db_table = 'webapi_webapitoken'
        verbose_name = _('Web API Token')
        verbose_name_plural = _('Web API Tokens')



"""Search-related testing utilities."""

from __future__ import unicode_literals

from contextlib import contextmanager

from django.core.management import call_command
from djblets.siteconfig.models import SiteConfiguration

from reviewboard.admin.siteconfig import load_site_config


def reindex_search():
    """Rebuild the search index."""
    call_command('rebuild_index', interactive=False)


@contextmanager
def search_enabled(on_the_fly_indexing=False):
    """Temporarily enable indexed search.

    Args:
        on_the_fly_indexing (bool, optional):
            Whether or not to enable on-the-fly indexing.
    """
    siteconfig = SiteConfiguration.objects.get_current()

    if on_the_fly_indexing:
        siteconfig.set('search_on_the_fly_indexing', True)

    siteconfig.set('search_enable', True)
    siteconfig.save()
    load_site_config()

    try:
        yield
    finally:
        if on_the_fly_indexing:
            siteconfig.set('search_on_the_fly_indexing', False)

        siteconfig.set('search_enable', False)
        siteconfig.save()

        load_site_config()



"""Authentication backend registry."""

from __future__ import unicode_literals

import logging
from warnings import warn

from django.conf import settings
from django.contrib.auth import get_backends
from django.utils.translation import ugettext_lazy as _
from djblets.registries.registry import (ALREADY_REGISTERED, LOAD_ENTRY_POINT,
                                         NOT_REGISTERED, UNREGISTER)

from reviewboard.accounts.backends.base import BaseAuthBackend
from reviewboard.accounts.backends.standard import StandardAuthBackend
from reviewboard.deprecation import RemovedInReviewBoard40Warning
from reviewboard.registries.registry import EntryPointRegistry


_enabled_auth_backends = []
_auth_backend_setting = None


class AuthBackendRegistry(EntryPointRegistry):
    """A registry for managing authentication backends."""

    entry_point = 'reviewboard.auth_backends'
    lookup_attrs = ('backend_id',)

    errors = {
        ALREADY_REGISTERED: _(
            '"%(item)r" is already a registered authentication backend.'
        ),
        LOAD_ENTRY_POINT: _(
            'Error loading authentication backend %(entry_point)s: %(error)s'
        ),
        NOT_REGISTERED: _(
            'No authentication backend registered with %(attr_name)s = '
            '%(attr_value)s.'
        ),
        UNREGISTER: _(
            '"%(item)r is not a registered authentication backend.'
        ),
    }

    def process_value_from_entry_point(self, entry_point):
        """Load the class from the entry point.

        If the class lacks a value for
        :py:attr:`~reviewboard.accounts.backends.base.BaseAuthBackend
        .backend_id`, it will be set as the entry point's name.

        Args:
           entry_point (pkg_resources.EntryPoint):
                The entry point.

        Returns:
            type:
            The :py:class:`~reviewboard.accounts.backends.base.BaseAuthBackend`
            subclass.
        """

        cls = entry_point.load()

        if not cls.backend_id:
            logging.warning('The authentication backend %r did not provide '
                            'a backend_id attribute. Setting it to the '
                            'entry point name ("%s")',
                            cls, entry_point.name)

            cls.backend_id = entry_point.name

        return cls

    def get_defaults(self):
        """Yield the authentication backends.

        This will make sure the standard authentication backend is always
        registered and returned first.

        Yields:
            type:
            The :py:class:`~reviewboard.accounts.backends.base.BaseAuthBackend`
            subclasses.
        """
        yield StandardAuthBackend

        for value in super(AuthBackendRegistry, self).get_defaults():
            yield value

    def unregister(self, backend_class):
        """Unregister the requested authentication backend.

        Args:
            backend_class (type):
                The class of the backend to unregister. This must be a subclass
                of :py:class:`~reviewboard.accounts.backends.base
                .BaseAuthBackend`.

        Raises:
            djblets.registries.errors.ItemLookupError:
                Raised when the class cannot be found.
        """
        self.populate()

        try:
            super(AuthBackendRegistry, self).unregister(backend_class)
        except self.lookup_error_class as e:
            logging.error('Failed to unregister unknown authentication '
                          'backend "%s".',
                          backend_class.backend_id)
            raise e

    def get_auth_backend(self, auth_backend_id):
        """Return the requested authentication backend, if it exists.

        Args:
            auth_backend_id (unicode):
                The unique ID of the
                :py:class:`~reviewboard.accounts.backends.base.BaseAuthBackend`
                class.

        Returns:
            type:
            The :py:class:`~reviewboard.accounts.backends.base.BaseAuthBackend`
            subclass, or ``None`` if it is not registered.
        """
        return self.get('backend_id', auth_backend_id)


def get_enabled_auth_backends():
    """Return all authentication backends being used by Review Board.

    The returned list contains every authentication backend that Review Board
    will try, in order.

    Returns:
        list of type:
        The list of registered
        :py:class:`~reviewboard.accounts.backends.base.BaseAuthBackend`
        subclasses.
    """
    global _enabled_auth_backends
    global _auth_backend_setting

    if (not _enabled_auth_backends or
        _auth_backend_setting != settings.AUTHENTICATION_BACKENDS):
        _enabled_auth_backends = []

        for backend in get_backends():
            if not isinstance(backend, BaseAuthBackend):
                warn('Authentication backends should inherit from '
                     'reviewboard.accounts.backends.base.BaseAuthBackend. '
                     'Please update %s.'
                     % backend.__class__,
                     RemovedInReviewBoard40Warning)

                for field, default in (('name', None),
                                       ('supports_registration', False),
                                       ('supports_change_name', False),
                                       ('supports_change_email', False),
                                       ('supports_change_password', False)):
                    if not hasattr(backend, field):
                        warn("Authentication backends should define a '%s' "
                             "attribute. Please define it in %s or inherit "
                             "from BaseAuthBackend."
                             % (field, backend.__class__),
                             RemovedInReviewBoard40Warning)

                        setattr(backend, field, False)

            _enabled_auth_backends.append(backend)

        _auth_backend_setting = settings.AUTHENTICATION_BACKENDS

    return _enabled_auth_backends


#: Registry instance for working with available authentication backends.
auth_backends = AuthBackendRegistry()
